# -*- coding: utf-8 -*-
"""
Polygons are assumed to be:
    - finite
    - simple (non self-intersecting)
    - non-explicitly closed
    - have no repeated points

"""


import numpy as np
from matplotlib.path import Path
import datastock as ds


from . import _class00_poly2d_check as _poly2d_check


__all__ = ['main']


# ###########################################################
# ###########################################################
#              main
# ###########################################################


def main(
    x0=None,
    x1=None,
    key=None,
    # options for edges
    dedge=None,
    dsurface=None,
):

    # ----------------
    # check inputs
    # ----------------

    # polygon formatting
    din = _poly2d_check.check(
        x0=x0,
        x1=x1,
        key=key,
        # options
        closed=True,
        clockwise=True,
    )

    # dedge, dsurface
    dedge, dsurface = _check(
        dedge=dedge,
        dsurface=dsurface,
    )

    # ----------------
    # sample edges
    # ----------------

    if dedge is not None:
        dout_edge = _edges(
            x0_closed=din['x0'],
            x1_closed=din['x1'],
            key=din['key'],
            # options
            **{k0: dedge.get(k0) for k0 in ['res', 'factor']},
        )
    else:
        dout_edge = {
            'x0': [],
            'x1': [],
        }

    # ----------------
    # sample surface
    # ----------------

    if dsurface is not None:
        dout_surf = _surface(
            x0_closed=din['x0'],
            x1_closed=din['x1'],
            key=din['key'],
            # options
            **{k0: dsurface.get(k0) for k0 in ['res', 'nb']},
        )
    else:
        dout_surf = {
            'x0': [],
            'x1': [],
        }

    # ----------------
    # combine
    # ----------------

    dout = {
        'x0': np.r_[dout_edge['x0'], dout_surf['x0']],
        'x1': np.r_[dout_edge['x1'], dout_surf['x1']],
        'edge_res': dout_edge.get('res'),
        'edge_factor': dout_edge.get('factor'),
        'surface_nb': dout_surf.get('nb'),
        'key': din['key'],
    }

    return dout


# ###########################################################
# ###########################################################
#              check main
# ###########################################################


def _check(
    dedge=None,
    dsurface=None,
):

    # --------------
    # dedge
    # --------------

    if dedge is not None:
        lk = ['res', 'factor']
        c0 = (
            isinstance(dedge, dict)
            and all([kk in lk for kk in dedge.keys()])
        )

        if not c0:
            lstr = [f"\t- {kk}" for kk in lk]
            msg = (
                "Arg dedge must be None or a dict with keys:\n"
                + "\n".join(lstr)
            )
            raise Exception(msg)

    # --------------
    # dsurface
    # --------------

    if dsurface is not None:
        lk = ['res', 'nb']
        c0 = (
            isinstance(dsurface, dict)
            and all([kk in lk for kk in dsurface.keys()])
        )

        if not c0:
            lstr = [f"\t- {kk}" for kk in lk]
            msg = (
                "Arg dsurface must be None or a dict with keys:\n"
                + "\n".join(lstr)
            )
            raise Exception(msg)

    return dedge, dsurface


# ###########################################################
# ###########################################################
#              Sample edges
# ###########################################################


def _edges(
    x0_closed=None,
    x1_closed=None,
    key=None,
    # options
    res=None,
    factor=None,
):

    # ----------------
    # check inputs
    # ----------------

    # options
    res, factor = _check_edges(
        res=res,
        factor=factor,
    )

    # ----------------
    # prepare
    # ----------------

    # get all lengths
    dx0 = x0_closed[1:] - x0_closed[:-1]
    dx1 = x1_closed[1:] - x1_closed[:-1]
    dist = np.hypot(dx0, dx1)

    # ----------------
    # determine samplung res
    # ----------------

    if isinstance(res, str):

        # get res
        if res == 'min':
            res = np.min(dist)
        elif res == 'max':
            res = np.max(dist)
        else:
            raise NotImplementedError()

        # adjust
        res = res * factor

    # ----------------
    # sample res
    # ----------------

    # nb of points to be inserted in each segment
    npts = np.round(dist / res, decimals=0).astype(int)

    # interpolation for x0
    out0 = np.concatenate([
        np.r_[x0_closed[ii]]
        if npts[ii] <= 1 else
        x0_closed[ii] + dx0[ii] * np.linspace(0, 1, npts[ii], endpoint=False)
        for ii in range(x0_closed.size-1)
    ])

    # interpolation for x1
    out1 = np.concatenate([
        np.r_[x1_closed[ii]]
        if npts[ii] <= 1 else
        x1_closed[ii] + dx1[ii] * np.linspace(0, 1, npts[ii], endpoint=False)
        for ii in range(x1_closed.size-1)
    ])

    # ----------------
    # return
    # ----------------

    dout = {
        'x0': out0,
        'x1': out1,
        'res': res,
        'factor': factor,
    }

    return dout


# ###########################################################
# ###########################################################
#              check edges
# ###########################################################


def _check_edges(
    res=None,
    factor=None,
):

    # -------------
    # res
    # -------------

    res = ds._generic_check._check_var(
        res, 'res',
        types=(str, float),
    )

    if isinstance(res, str):
        res = ds._generic_check._check_var(
            res, 'res',
            types=str,
            default='min',
            allowed=['min', 'max'],
        )

    else:
        res = ds._generic_check._check_var(
            res, 'res',
            types=float,
            sign='>0',
        )

    # -------------
    # factor
    # -------------

    if isinstance(res, str):
        factor = float(ds._generic_check._check_var(
            factor, 'factor',
            types=(int, float),
            default=1,
            sign='>0',
        ))

    else:
        factor = None

    return res, factor


# ###########################################################
# ###########################################################
#              Sample surface
# ###########################################################


def _surface(
    x0_closed=None,
    x1_closed=None,
    key=None,
    # options
    res=None,
    nb=None,
):

    # ----------------
    # check inputs
    # ----------------

    # Dx0
    x0_min = x0_closed.min()
    x0_max = x0_closed.max()
    Dx0 = (x0_max - x0_min)

    # Dx1
    x1_min = x1_closed.min()
    x1_max = x1_closed.max()
    Dx1 = (x1_max - x1_min)

    # options
    nb = _check_surfaces(
        res=res,
        nb=nb,
        key=key,
        Dx0=Dx0,
        Dx1=Dx1,
    )

    # ----------------
    # prepare
    # ----------------

    # out0
    dx0 = Dx0 / nb[0]
    out0 = np.linspace(x0_min + dx0/2, x0_max - dx0/2, nb[0])

    # out1
    dx1 = Dx1 / nb[1]
    out1 = np.linspace(x1_min + dx1/2, x1_max - dx1/2, nb[1])

    # ----------------
    # Check in polygon
    # ----------------

    pts0 = np.repeat(out0[:, None], out1.size, axis=1).ravel()
    pts1 = np.repeat(out1[None, :], out1.size, axis=0).ravel()

    pp = Path(np.array([x0_closed, x1_closed]).T)
    ind = pp.contains_points(np.array([pts0, pts1]).T)

    # ----------------
    # return
    # ----------------

    dout = {
        'x0': pts0[ind],
        'x1': pts1[ind],
        'nb': nb,
    }

    return dout


# ###########################################################
# ###########################################################
#              check surfaces
# ###########################################################


def _check_surfaces(
    res=None,
    nb=None,
    key=None,
    Dx0=None,
    Dx1=None,
):

    # -------------
    # res vs nb
    # -------------

    lc = [
        nb is not None and res is not None,
        nb is None and res is None,
    ]
    if any(lc):
        msg = (
            "Polygon '{key}' surface sampling, please provide res xor nb!\n"
            f"\t- res = {res} \n"
            f"\t- nb = {nb}\n"
        )
        raise Exception(msg)

    # -------------
    # res
    # -------------

    if res is not None:

        if np.isscalar(res):
            res = [res, res]

        res = ds._generic_check._check_var_iter(
            res, 'res',
            types=(list, tuple),
            types_iter=(int, float),
            size=2,
        )

        res = tuple([float(rr) for rr in res])

        nb = [np.ceil(Dx0/res[0]), np.ceil(Dx1/res[1])]

    # -------------
    # nb
    # -------------

    if np.isscalar(nb):
        nb = [nb, nb]

    nb = ds._generic_check._check_var_iter(
        nb, 'nb',
        types=(list, tuple),
        types_iter=(int, float),
        size=2,
    )

    nb = tuple([int(nn) for nn in nb])

    return nb


# -*- coding: utf-8 -*-


import numpy as np
import datastock as ds


# ##################################################################
# ##################################################################
#                       Diagnostics
# ##################################################################


def _isconvex(coll=None, keys=None):

    # ------------
    # check inputs

    if isinstance(keys, str):
        keys = [keys]

    lap = list(coll.dobj.get('aperture', {}).keys())
    lfilt = list(coll.dobj.get('filter', {}).keys())
    lcryst = list(coll.dobj.get('crystal', {}).keys())
    lgrat = list(coll.dobj.get('grating', {}).keys())

    lop = lap + lfilt + lcryst + lgrat

    keys = ds._generic_check._check_var_iter(
        keys, 'keys',
        types=(list, tuple, np.ndarray),
        types_iter=str,
        allowed=lop,
    )

    # ------
    # loop

    isconvex = []
    for k0 in keys:

        # get class
        if k0 in lap:
            cls = 'aperture'
        elif k0 in lfilt:
            cls = 'filter'
        elif k0 in lcryst:
            cls = 'crystal'
        else:
            cls = 'grating'

        # dgeom
        curve_r = coll.dobj[cls][k0]['dgeom'].get('curve_r')
        if curve_r is None:
            convex = False
        else:
            if any([np.isfinite(rr) and rr > 0. for rr in curve_r]):
                convex = True
            else:
                convex = False
        isconvex.append(convex)

    return isconvex




import numpy as np
import datastock as ds


# ##################################################################
# ##################################################################
#                   plot text fror ref on axes
# ##################################################################


def plot_text(
    coll=None,
    kax=None,
    ax=None,
    ref=None,
    group=None,
    ind=None,
    lkeys=None,
    nmax=None,
    color_dict=None,
    bstr_dict=None,
):

    # ------------------
    # Get list of data

    lk0 = ['index'] + [
        k0 for k0, v0 in coll._ddata.items()
        if v0['ref'] == (ref,)
    ]

    if isinstance(lkeys, str):
        lkeys = [lkeys]
    lk0 = ds._generic_check._check_var_iter(
        lkeys, 'lkeys',
        default=lk0,
        types=list,
        types_iter=str,
        allowed=lk0,
    )

    # ------------------
    # Get list of data

    if bstr_dict is None:
        bstr_dict = '{}'
    if isinstance(bstr_dict, str):
        bstr_dict = {k0: bstr_dict for k0 in lk0}
    c0 = (
        isinstance(bstr_dict, dict)
        and all([
            k0 in bstr_dict.keys()
            and isinstance(bstr_dict.get(k0), str)
            for k0 in lk0
        ])
    )
    if not c0:
        msg = (
            "Arg bstr_dict must be a dict with keys:\n"
            f"\t- keys: {lk0}\n"
            f"\t- Provided: {bstr_dict}"
        )
        raise Exception(msg)

    # ------------------
    # Compute placements

    nx = nmax + 1
    ny = len(lk0) + 1
    x = np.linspace(0, 1, nx + 1)
    y = np.linspace(0, 1, ny + 1)

    # ----------------
    # Plot fixed text

    ax.text(
        np.mean(x),
        y[-1],
        f'{ref}',
        horizontalalignment='center',
        verticalalignment='top',
        transform=ax.transAxes,
        size=11,
        color='k',
        fontweight='bold',
    )

    for ii, k0 in enumerate(lk0):
        ax.text(
            x[0],
            y[-(ii + 2)],
            k0,
            horizontalalignment='left',
            verticalalignment='top',
            transform=ax.transAxes,
            size=8,
            color='k',
            fontweight='bold',
        )

    # ----------------
    # Plot mobile text

    for ii, k0 in enumerate(lk0):
        for jj in range(nmax):
            if k0 == 'index':
                datastr = ind
                data = k0
            else:
                datastr = coll._ddata[k0]['data'][ind]
                data = k0
            ht = ax.text(
                x[jj + 1],
                y[-(ii + 2)],
                data,
                horizontalalignment='left',
                verticalalignment='top',
                transform=ax.transAxes,
                size=8,
                color=color_dict[group][jj],
                fontweight='bold',
            )

            coll.add_mobile(
                key=f'txt-{ref}-{k0}-{jj}',
                handle=ht,
                ref=(ref,),
                data=data,
                dtype='txt',
                bstr=bstr_dict[k0],
                ax=kax,
                ind=jj,
            )



# -*- coding: utf-8 -*-
"""
Created on Sat Jun 29 20:39:20 2024

@author: dvezinet
"""


import json


import numpy as np
import datastock as ds


# #################################################################
# #################################################################
#          Main
# #################################################################


def main(
    # ---------------
    # input from tofu
    pfe=None,
    coll=None,
    fname=None,
    returnas=None,
):

    # ----------------
    # check inputs
    # --------------

    dout, returnas = _check(
        coll=coll,
        pfe=pfe,
        returnas=returnas,
    )

    # ----------------
    # get file content
    # ----------------

    # ------------
    # fill optics

    for kcls, vcls in dout.items():

        if kcls == 'diagnostic':
            continue

        for k1, v1 in vcls.items():
            _add_optics(
                coll=coll,
                which=kcls,
                din=v1,
                fname=fname,
            )

    # ---------------
    # add diagnostic

    _add_diagnostic(
        coll=coll,
        din=dout['diagnostic'],
    )

    if returnas is True:
        return coll


# #################################################################
# #################################################################
#          check
# #################################################################


def _check(
    coll=None,
    pfe=None,
    returnas=None,
):

    # --------------------
    # load dict from file
    # --------------------

    with open(pfe, 'r') as fn:
        dout = json.load(fn)

    # --------------------
    # check content
    # --------------------

    c0 = (
        isinstance(dout, dict)
        and all([ss in dout.keys() for ss in ['diagnostic', 'camera']])
    )
    if not c0:
        msg = (
            "File does not seem to hold a tofu-compatible diagnostic!\n"
            f"\t- Provided: {pfe}\n"
        )
        raise Exception(msg)

    # -------------
    # returnas
    # -------------

    returnas = ds._generic_check._check_var(
        returnas, 'returnas',
        types=bool,
        default=True,
    )

    return dout, returnas


# #################################################################
# #################################################################
#          add_optics
# #################################################################


def _add_optics(
    coll=None,
    which=None,
    din=None,
    fname=None,
):

    # ----------------
    # key
    # ----------------

    lout = list(coll.dobj.get(which, {}).keys())
    if din['key'] in lout:
        din['key'] = f"{fname}_{din['key']}"

    # ----------------
    # dgeom
    # ----------------

    dgeom = _get_dgeom(din)

    # ----------------
    # dmat
    # ----------------

    dmat = _get_dmat(din)

    # ----------------
    # add to coll
    # ----------------

    if which == 'aperture':
        coll.add_aperture(
            key=din['key'],
            **dgeom,
        )

    else:
        if which == 'camera':
            func = f"add_camera_{dgeom['nd']}"
            del dgeom['nd']
        elif which in ['aperture', 'filter', 'crystal', 'grating']:
            func = f'add_{which}'
        else:
            msg = f"Unknow optics class: {which}"
            raise Exception(msg)

        # add
        getattr(coll, func)(
            key=din['key'],
            dgeom=dgeom,
            dmat=dmat,
        )

    return


# #################################################################
# #################################################################
#          _get_dgeom
# #################################################################


def _get_dgeom(din=None):

    # --------------
    # prepare
    # --------------

    larray = [
        'cent',
        'cents_x0', 'cents_x1',
        'cents_x', 'cents_y', 'cents_z',
        'nin', 'e0', 'e1',
        'nin_x', 'nin_y', 'nin_z',
        'e0_x', 'e0_y', 'e0_z',
        'e1_x', 'e1_y', 'e1_z',
        'outline_x0', 'outline_x1',
        'poly_x', 'poly_y', 'poly_z',
    ]
    lex = ['type', 'extenthalf', 'shape', 'parallel']

    # --------------
    # loop
    # --------------

    dgeom = {}
    for k0, v0 in din['dgeom'].items():

        if k0 in lex:
            continue

        if k0 in larray:
            v0 = np.asarray(v0['data'])
        elif isinstance(v0, dict):
            v0 = v0['data']

        dgeom[k0] = v0

    return dgeom


# #################################################################
# #################################################################
#          _get_dmat
# #################################################################


def _get_dmat(din=None):

    dmat = din.get('dmat')

    return dmat


# #################################################################
# #################################################################
#          _get_diagnostic
# #################################################################



def _add_diagnostic(
    coll=None,
    din=None,
):

    # ----------------
    # prepare doptics
    # ----------------

    doptics = din['doptics']

    # --------------
    # add
    # --------------

    coll.add_diagnostic(
        key=din['key'],
        doptics=doptics,
        compute=False,
    )

    # force camera order (to maintain arrays order)
    coll._dobj['diagnostic'][din['key']]['camera'] = din['camera']

    # ------------------
    # add computed data
    # -------------------

    lcam = din['camera']

    # -----------
    # los

    for kcam in lcam:

        # add ray
        klos = doptics[kcam].get('los_key')
        if klos is not None:
            coll.add_rays(
                key=klos,
                # start
                start_x=doptics[kcam]['los_x_start']['data'],
                start_y=doptics[kcam]['los_y_start']['data'],
                start_z=doptics[kcam]['los_z_start']['data'],
                # pts
                pts_x=doptics[kcam]['los_x_end']['data'],
                pts_y=doptics[kcam]['los_y_end']['data'],
                pts_z=doptics[kcam]['los_z_end']['data'],
                # angles
                alpha=doptics[kcam]['los_alpha']['data'],
                dalpha=doptics[kcam]['los_dalpha']['data'],
                dbeta=doptics[kcam]['los_dbeta']['data'],
            )

            # adjust ref to match camera
            ref = tuple(
                [coll.dobj['rays'][klos]['ref'][0]]
                + list(coll.dobj['camera'][kcam]['dgeom']['ref'])
            )
            coll._dobj['rays'][klos]['ref'] = ref

        # store in diag
        coll._dobj['diagnostic'][din['key']]['doptics'][kcam]['los'] = klos

    # -----------
    # etendue

    for kcam in lcam:

        # add data
        ketend = doptics[kcam].get('etendue_key')
        etend_type = doptics[kcam].get('etend_type')
        if ketend is not None:
            ref = coll.dobj['camera'][kcam]['dgeom']['ref']
            coll.add_data(
                key=ketend,
                data=doptics[kcam]['etendue']['data'],
                units=doptics[kcam]['etendue']['units'],
                ref=ref,
            )

        # store in diag
        coll._dobj['diagnostic'][din['key']]['doptics'][kcam]['etendue'] = ketend
        coll._dobj['diagnostic'][din['key']]['doptics'][kcam]['etend_type'] = etend_type

    # -----------
    # vos
    # -----------

    for kcam in lcam:

        # --------
        # add data

        for kp in ['pcross_x0', 'pcross_x1', 'phor_x0', 'phor_x1']:

            kpi = doptics[kcam][kp]['key']
            ref = doptics[kcam][kp]['ref']

            # add ref if needed
            if ref is not None:
                lr = [rr for rr in ref if rr not in coll.dref.keys()]
                for rr in lr:
                    ii = ref.index(rr)
                    coll.add_ref(
                        key=rr,
                        size=np.array(doptics[kcam][kp]['data']).shape[ii],
                    )

            # add data
            if kpi is not None:
                coll.add_data(
                    key=kpi,
                    data=doptics[kcam][kp]['data'],
                    units=doptics[kcam][kp]['units'],
                    ref=tuple(ref),
                )

        # -----------------
        # store vos in diag

        pcross0 = doptics[kcam].get('pcross_x0', {}).get('key')
        pcross1 = doptics[kcam].get('pcross_x1', {}).get('key')
        phor0 = doptics[kcam].get('phor_x0', {}).get('key')
        phor1 = doptics[kcam].get('phor_x1', {}).get('key')
        dphi = doptics[kcam].get('dphi', {}).get('data')

        if pcross0 is not None:
            pcross = (pcross0, pcross1)
            phor = (phor0, phor1)
            dphi = np.array(dphi)
        else:
            pcross = None
            phor = None
            dphi = None

        # -----------------
        # store phor in diag

        coll._dobj['diagnostic'][din['key']]['doptics'][kcam]['dvos'] = {
            'pcross': pcross,
            'phor': phor,
            'dphi': dphi,
        }

    return


# -*- coding: utf-8 -*-
"""
Created on Thu May 30 13:39:45 2024

@author: dvezinet
"""


import itertools as itt


import numpy as np
import datastock as ds


# ##################################################################
# ##################################################################
#                   main
# ##################################################################


def main(
    coll=None,
    key=None,
    key_data=None,
    key_cam=None,
    flat=None,
):

    # ------------
    # check inputs

    key, key_data, key_cam, key_cam_equi, is2d, stack, ref, flat = _check(
        coll=coll,
        key=key,
        key_data=key_data,
        key_cam=key_cam,
        flat=flat,
    )

    # ------------
    # prepare

    if is2d:
        ax0 = ref.index(None)
        ax1 = len(ref) - 1 - ref[::-1].index(None)
        if flat:
            ldata = []
            for k0 in key_data:
                sh = list(coll.ddata[k0]['data'].shape)
                size = sh[ax0] * sh[ax1]
                sh = tuple(np.r_[sh[:ax0], size, sh[ax1+1:]].astype(int))
                ldata.append(coll.ddata[k0]['data'].reshape(sh))
            axis = ax0
        else:
            axis = ax0 if stack == 'horizontal' else ax1
            ldata = [coll.ddata[k0]['data'] for k0 in key_data]
            lsize = [dd.shape[axis] for dd in ldata]
            if len(set(lsize)) != 1:
                msg = (
                    "Data for diag '{key}' cannot be stacked {stack}:\n"
                    f"\t- shapes: {[dd.shape for dd in ldata]}\n"
                    f"\t- axis: {axis}"
                )
                raise Exception(msg)

    else:
        ldata = [coll.ddata[k0]['data'] for k0 in key_data]
        axis = ref.index(None)

    # ------------
    # concatenate

    data = np.concatenate(tuple(ldata), axis=axis)
    units = coll.ddata[key_data[0]]['units']

    # dind
    i0 = 0
    dind = {}
    for ii, k0 in enumerate(key_data):
        npix = ldata[ii].shape[axis]
        ind = i0 + np.arange(0, npix)
        dind[k0] = ind
        i0 += npix

    return {
        'data': data,
        'keys': key_data,
        'keys_cam': key_cam,
        'keys_cam_equi': key_cam_equi,
        'units': units,
        'ref': ref,
        'axis': axis,
        'flat': flat,
        'dind': dind,
    }


# ##################################################################
# ##################################################################
#                   check
# ##################################################################


def _check(
    coll=None,
    key=None,
    key_data=None,
    key_cam=None,
    flat=None,
):

    # ---------------
    # key and key_cam
    # ---------------

    key, key_cam = coll.get_diagnostic_cam(key=key, key_cam=key_cam)
    # spectro = coll.dobj['diagnostic'][key]['spectro']
    is2d = coll.dobj['diagnostic'][key]['is2d']
    stack = coll.dobj['diagnostic'][key]['stack']

    # ---------------------------
    # get list of equivalent camera series
    # ---------------------------

    wdiag = 'diagnostic'
    wcam = 'camera'
    ldiag_equi = [
        kdiag
        for kdiag, vdiag in coll.dobj.get(wdiag, {}).items()
        if kdiag != key
        and len(vdiag['camera']) == len(coll.dobj[wdiag][key]['camera'])
        and all([
            coll.dobj[wcam][vdiag['camera'][ii]]['dgeom']['shape']
            == coll.dobj[wcam][kcam]['dgeom']['shape']
            for ii, kcam in enumerate(coll.dobj[wdiag][key]['camera'])
        ])
    ]

    # ---------------------------
    # get list of equivalent camera series
    # ---------------------------

    dcam_equi = {
        kdiag: [
            coll.dobj[wdiag][kdiag]['camera'][ii]
            for ii, kcam in enumerate(coll.dobj[wdiag][key]['camera'])
            if kcam in key_cam
        ]
        for kdiag in ldiag_equi
    }

    # ------------
    # key_data
    # ------------

    # ----------
    # key_data

    if isinstance(key_data, str):

        if key_data in coll.ddata.keys():
            key_data = [key_data]

        else:
            lok = coll.dobj['diagnostic'][key]['signal']
            lok_equi = list(itt.chain.from_iterable([
                coll.dobj['diagnostic'][kdiag]['signal']
                for kdiag in ldiag_equi
            ]))

            if lok is None:
                lok = []
            if lok_equi is None:
                lok_equi = []

            key_data = ds._generic_check._check_var(
                key_data, 'key_data',
                types=str,
                allowed=lok + lok_equi,
            )

            key_data = coll.dobj['synth sig'][key_data]['data']

    # ------------------------
    # basic consistency check
    # ------------------------

    c0 = (
        isinstance(key_data, list)
        and all([
            isinstance(kk, str) and kk in coll.ddata.keys()
            for kk in key_data
        ])
    )
    if not c0:
        msg = (
            "Arg key_data must be a list of valid data keys!\n"
            f"Provided: {key_data}"
        )
        raise Exception(msg)

    # -----------------------
    # check cameras validity
    # -----------------------

    lcam = [coll.ddata[k0]['camera'] for k0 in key_data]

    lc = (
        [all([kcam in key_cam for kcam in lcam])]
        + [
            all([kcam in dcam_equi[kdiag] for kcam in lcam])
            for kdiag in ldiag_equi
        ]
    )

    if not any(lc):
        lstr = (
            [f"\t- {key_cam}"]
            + [f"\t- {dcam_equi[kdiag]}" for kdiag in ldiag_equi]
        )
        lstr2 = [f"\t- '{k0}': '{lcam[ii]}'" for ii, k0 in enumerate(key_data)]
        msg = (
            "All data provided must be associated to one "
            "of the following cameras series:\n"
            + "\n".join(lstr)
            + "But the provided data has:\n"
            + "\n".join(lstr2)
        )
        raise Exception(msg)

    # ---------------------
    # check cameras unicity
    # ---------------------

    # check unicity of cameras
    if len(set(lcam)) > len(lcam):
        msg = (
            f"Non-unique camera references for diag '{key}'\n:"
            f"\t- key_data = {key_data}\n"
            f"\t- lcam     = {lcam}\n"
        )
        raise Exception(msg)

    # ---------------------
    # is_equi
    # ---------------------

    if lc[0]:

        key_cam = [k0 for k0 in key_cam if k0 in lcam]
        key_cam_equi = None
        key_cam_ref = [coll.dobj[wcam][kk]['dgeom']['ref'] for kk in key_cam]

    else:
        kdiag = ldiag_equi[lc.index(True) - 1]
        key_cam_equi = [k0 for k0 in dcam_equi[kdiag] if k0 in lcam]
        key_cam = [
            k0 for ii, k0 in enumerate(key_cam)
            if dcam_equi[kdiag][ii] in lcam
        ]

        key_cam_ref = [coll.dobj[wcam][kk]['dgeom']['ref'] for kk in key_cam_equi]

    # ------------
    # re-order data
    # -------------

    if key_cam_equi is None:
        key_data = [key_data[lcam.index(k0)] for k0 in key_cam]
    else:
        key_data = [key_data[lcam.index(k0)] for k0 in key_cam_equi]

    # check uniformity of ref dimensions
    lref = [coll.ddata[k0]['ref'] for k0 in key_data]
    if any([len(ref) != len(lref[0]) for ref in lref[1:]]):
        msg = (
            f"Non uniform refs length for data in diag '{key}':\n"
            f"\t- key_data = {key_data}\n"
            f"\t- lref = {lref}\n"
        )
        raise Exception(msg)

    # -------------
    # ref and axis
    # -------------

    laxcam = [
        [ref.index(rr) for rr in key_cam_ref[ii]]
        for ii, ref in enumerate(lref)
    ]

    # check all 2d or all 1d (future: flatten if needed)
    if any([len(ax) != len(laxcam[0]) for ax in laxcam[1:]]):
        msg = (
            f"Non-uniform camera concatenation axis for diag '{key}':\n"
            f"\t- key_data: {key_data}\n"
            f"\t- laxcam:   {laxcam}"
        )
        raise Exception(msg)

    # concatenation axis uniformity
    laxcam = np.array(laxcam)
    if not np.allclose(laxcam, laxcam[0:1, :]):
        msg = (
            f"Non-uniform axis for concatenation for diag '{key}':\n"
            f"\t- laxcam: {laxcam}"
        )
        raise Exception(msg)
    axcam = laxcam[0]

    # consistency vs is2d
    if len(axcam) != 1 + is2d:
        msg = (
            f"ref not consistent with is2d for diag '{key}':\n"
            f"\t- is2d: {is2d}\n"
            f"\t- axcam: {axcam}\n"
        )
        raise Exception(msg)

    # get unique ref and set concatenation axis to None
    ref = list(lref[0])
    for ii in axcam:
        ref[ii] = None

    # --------
    # flat
    # --------

    flat = ds._generic_check._check_var(
        flat, 'flat',
        types=bool,
        default=is2d,
    )

    return key, key_data, key_cam, key_cam_equi, is2d, stack, ref, flat




# Standard
import numpy as np
import matplotlib.pyplot as plt


# tofu-specific
import tofu.physics_tools as tfpt


#######################################################
#
#     Setup and Teardown
#
#######################################################


def setup_module():
    pass


def teardown_module():
    pass


#######################################################
#
#     checking Distributions
#
#######################################################


class Test01_Distributions():

    @classmethod
    def setup_class(cls):
        pass

    def setup_method(self):
        pass

    def teardown_method(self):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def test01_maxwellian(self):

        E = np.linspace(0.1, 20, 100) * 1e3
        kTe = np.r_[0.1, 1, 10, 100] * 1e3

        # single
        dout = tfpt.get_maxwellian(
            kTe_eV=kTe[0],
            energy_eV=E,
        )
        assert isinstance(dout, dict)

        # arrays
        dout = tfpt.get_maxwellian(
            kTe_eV=kTe[None, :],
            energy_eV=E[:, None],
        )
        assert isinstance(dout, dict)

        # wavelength
        dout = tfpt.get_maxwellian(
            kTe_eV=kTe[None, :],
            velocity_ms=np.linspace(1, 5, 10)[:, None]*1e6,
        )
        assert isinstance(dout, dict)


#######################################################
#
#     checking Runaways
#
#######################################################


class Test02_Runaways():

    @classmethod
    def setup_class(cls):
        pass

    def setup_method(self):
        pass

    def teardown_method(self):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def test01_convert(self):

        # from gamma
        beta = tfpt.runaways.convert_momentum_velocity_energy(
            gamma=[1, 2, 3],
        )['beta']['data']
        assert np.all((beta >= 0.) & (beta <= 1.))

        # from momentum normalized
        gamma = tfpt.runaways.convert_momentum_velocity_energy(
            momentum_normalized=10,
        )['gamma']['data']
        assert np.all(gamma >= 1.)

        # from kinetic energy
        dout = tfpt.runaways.convert_momentum_velocity_energy(
            energy_kinetic_eV=(1e3, 10e3),
        )
        assert isinstance(dout, dict)
        assert np.all(dout['velocity_ms']['data'] > 0.)
        assert np.all(dout['velocity_ms']['data'] < 3e8)

        # from velocity
        dout = tfpt.runaways.convert_momentum_velocity_energy(
            velocity_ms=1e6,
        )
        assert isinstance(dout, dict)

    def test02_electric_fields(self):
        dout = tfpt.runaways.get_critical_dreicer_electric_fields(
            ne_m3=np.r_[1e19, 1e20][None, :],
            kTe_eV=np.r_[1, 2, 3][:, None]*1e3,
            lnG=20,
        )
        assert isinstance(dout, dict)

    def test03_growth_source_terms(self):
        dout = tfpt.runaways.get_growth_source_terms(
            ne_m3=np.r_[1e19, 1e20][None, :],
            lnG=15,
            Epar_Vm=1,
            kTe_eV=np.r_[1, 2, 3][:, None]*1e3,
            Zeff=2,
        )
        assert isinstance(dout, dict)

    def test04_normalized_momentum_distribution(self):

        # case with both avalanche and dreicer
        pp = np.linspace(0.1, 10, 100)
        ne_m3 = np.r_[0.1, 1, 10, 100]*1e19
        Epar = 1
        Emax = 10e6

        # compute
        dout = tfpt.runaways.get_normalized_momentum_distribution(
            momentum_normalized=pp[:, None],
            ne_m3=ne_m3[None, :],
            Zeff=2.,
            electric_field_par_Vm=Epar,
            energy_kinetic_max_eV=Emax,
            lnG=None,
            sigmap=1.,
        )
        assert isinstance(dout, dict)

    def test05_emission_thick_anisotropy(self):
        E = np.r_[1, 10, 100] * 1e3
        gamma = tfpt.runaways.convert_momentum_velocity_energy(
            energy_kinetic_eV=E,
        )['gamma']['data']
        anis = tfpt.runaways.emission.get_xray_thick_anisotropy(
            gamma=gamma[None, :],
            costheta=np.linspace(-1, 1, 100)[:, None],
        )
        assert isinstance(anis, np.ndarray) and np.all(anis > 0)

    def test06_emission_get_xray_thick_dcross_ei(self):
        E_re = np.r_[1, 10, 100] * 1e3
        E_ph = np.linspace(1, 20, 50) * 1e3
        dout = tfpt.runaways.emission.get_xray_thick_dcross_ei(
            E_re_eV=E_re[None, :],
            E_ph_eV=E_ph[:, None],
            atomic_nb=13,
            adjust=True,
            costheta=None,
            return_intermediates=True,
        )
        assert isinstance(dout, dict)

    def test07_plot_xray_thick_dcross_ei_vs_Salvat(self):
        dax = tfpt.runaways.emission.plot_xray_thick_dcross_ei_vs_Salvat()
        plt.close('all')
        assert isinstance(dax, dict)



"""
Getting started: 5 minutes tutorial for `tofu`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

This is a tutorial that aims to get a new user a little familiar with tofu's
structure.
"""

# The following imports matplotlib, preferably using a
# backend that allows the plots to be interactive (Qt5Agg).
import numpy as np

###############################################################################
# We start by loading `tofu`. You might see some warnings at this stage since
# optional modules for `tofu` could
# be missing on the machine you are working on. This can be ignored safely.

import tofu as tf


###############################################################################
# We can now create our first configuration.
# In `tofu` speak, a configuration is the geometry of the device and its
# structures. `tofu` provides pre-defined ones for your to try, so we're going
# to do just that:

configB2 = tf.load_config("B2")

###############################################################################
# The configuration can easily be visualized using the `.plot()` method:

configB2.plot()

###############################################################################
# Since `tofu` is all about tomography, let's create a 1D camera and plot its
# output.

cam1d = tf.geom.utils.create_CamLOS1D(
    config=configB2,
    pinhole=[3.4, 0, 0],
    sensor_nb=100,
    focal=0.1,
    sensor_size=0.1,
    orientation=[np.pi, 0, 0],
    Name="",
    Exp="",
    Diag="",
)

# interactive plot
cam1d.plot_touch()

###############################################################################
# The principle is similar for 2D cameras.

cam2d = tf.geom.utils.create_CamLOS2D(
    config=configB2,
    pinhole=[3.4, 0, 0],
    sensor_nb=100,
    focal=0.1,
    sensor_size=0.1,
    orientation=[np.pi, 0, 0],
    Name="",
    Exp="",
    Diag="",
)
cam2d.plot_touch()

###############################################################################
# What comes next is up to you!
# You could now play with the function parameters (change the cameras
# direction, refinement, aperture),
# with the plots (many are interactive) or create you own tomographic
# configuration.




import os


import numpy as np
import scipy.constants as scpct
import scipy.interpolate as scpinterp
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import datastock as ds


from .. import _utils


# ##############################################################
# ##############################################################
#                        DEFAULTS
# ##############################################################


_PATH_HERE = os.path.dirname(__file__)


# ##############################################################
# ##############################################################
#               Bremsstrahlung anisotropy factor
# ##############################################################


def anisotropy(
    gamma=None,
    costheta=None,
):
    """ Return the anisotropic factor (unitless)

    Depends on:
        - gamma: thelorentz factor of the Runaway electron
        - costheta: angle of observation relative to electron direction

    ref:
    [1] Pandya et al., Physica Scripta 93, no. 11 (November 1, 2018): 115601

    """

    # -----------
    # check inputs
    # -----------

    gamma, costheta = _check_anisotropy(
        gamma=gamma,
        costheta=costheta,
    )

    # -----------
    # prepare
    # -----------

    # gamma => beta
    beta = _utils.convert_momentum_velocity_energy(
        gamma=gamma,
    )['beta']['data']

    # -----------
    # compute
    # -----------

    # anisotropy of cross-section
    anis = (
        (3/8) * (1 + ((costheta - beta) / (1 - beta * costheta))**2)
        / (gamma**2 * (1 - beta * costheta)**2)
    )

    return anis


def _check_anisotropy(
    gamma=None,
    costheta=None,
):

    dout, shape = ds._generic_check._check_all_broadcastable(
        gamma=gamma,
        costheta=costheta,
    )

    return dout['gamma'], dout['costheta']


# ##############################################################
# ##############################################################
#            Differencial Bremsstrahlung cross-section
# ##############################################################


def dcross_ei(
    E_re_eV=None,
    E_ph_eV=None,
    atomic_nb=None,
    adjust=None,
    # optional anistropy
    costheta=None,
    # options
    return_intermediates=None,
):
    """ Return the energy-dependent HXR generation cros-section

    Considers electron-ion bremsstrahlung

    To be multiplied by the anisotropy factor
        => see get_anisotropy_factor()

    Sources:
        [1] Nocente et al., Nuclear Fusion 57, no. 7 (July 1, 2017): 076016.
        [2] Salvat et al., Nuclear Instruments and Methods in Physics Research
            Section B: Beam Interactions with Materials and Atoms 63,
            no. 3 (February 1992): 255–69
    """

    # --------------------
    # check inputs
    # --------------------

    (
        E_re_eV, E_ph_eV,
        atomic_nb, adjust,
        costheta,
        return_intermediates,
    ) = _check_dcross_ei(
        E_re_eV=E_re_eV,
        E_ph_eV=E_ph_eV,
        atomic_nb=atomic_nb,
        adjust=adjust,
        # optional anistropy
        costheta=costheta,
        # options
        return_intermediates=return_intermediates,
    )

    # --------------------
    # Load tabulated data
    # --------------------

    # load screening radius
    fname = "RE_HXR_CrossSection_ScreeningRadius_Salvat.csv"
    pfe = os.path.join(_PATH_HERE, fname)
    Z_R, RZ3a0 = np.loadtxt(pfe, delimiter=',').T

    fname = "RE_HXR_ElectronElectron_Salvat.csv"
    pfe = os.path.join(_PATH_HERE, fname)
    Z_eta, eta_inf = np.loadtxt(pfe, delimiter=',').T

    # -----------------------------------
    # cross-section (without anisotropy)
    # ----------------------------------

    # -------------------
    # prepare constants

    # hbar (J.s => eV.s)
    hbar_eVs = scpct.hbar / scpct.e

    # rest energy of electron (J => eV)
    mc2_eV = scpct.m_e * scpct.c**2 / scpct.e

    # mc (eV.s/m)
    mc_eVsm = mc2_eV / scpct.c

    # fine structure constant (adim.)
    alpha = scpct.alpha

    # a0 = bohr radius (m)
    a0 = 5.291772e-11

    # screening radius (should be tabulate from fig. 4), m
    # R = 0.81 * a0 / Z**(1/3)
    Rz3a0 = scpinterp.interp1d(
        np.round(Z_R),
        RZ3a0,
        kind='linear',
    )(atomic_nb)
    R = Rz3a0 * a0 / atomic_nb**(1/3)

    # high-energy coulomb correction (adim.)
    aa = (alpha * atomic_nb)**2
    fc = aa * np.sum([1./(nn * (nn**2 + aa)) for nn in range(1, 101)])

    # log(R * mc/hbar)   adim
    logRmchb = np.log(R * mc_eVsm / hbar_eVs)

    # ----------------------
    # prepare (nEe,) vectors

    # should tabulate eta_inf vs Z from graph (cf. fig. 5), adim.
    # eta_inf = 1.158
    eta_inf = scpinterp.interp1d(
        np.round(Z_eta),
        eta_inf,
        kind='linear',
    )(atomic_nb)

    # eta
    eta = (E_re_eV / mc2_eV)**0.8 / ((E_re_eV / mc2_eV)**0.8 + 2.43) * eta_inf

    # correction term at low energies adim
    F2 = (
        (2.04 + 9.09 * alpha * atomic_nb)
        * (mc2_eV**2 / (E_re_eV * (E_re_eV + mc2_eV)))
        ** (1.26 - 0.93 * alpha * atomic_nb)
    )

    # useful for q0 = minimum momentum transfer (eV.s/m)
    gamma = _utils.convert_momentum_velocity_energy(
        energy_kinetic_eV=E_re_eV,
    )['gamma']['data']

    # --------------------------
    # prepare (nEe, nEph) arrays

    # reduced energy of the photon (adim.)
    eps = E_ph_eV / (E_re_eV + mc2_eV)
    eps[eps > 1] = np.nan
    epsd = (E_re_eV - 5 * mc2_eV) / (E_re_eV + mc2_eV)

    # turn off fc when eps > epsd
    theta = eps < epsd

    # q0 = minimum momentum transfer (eV.s/m)
    q0 = (mc_eVsm / (2 * gamma)) * eps / (1 - eps)

    # adim
    bb = R * q0 / hbar_eVs

    # Phi1
    Phi1 = 2 - 2*np.log(1 + bb**2) - 4*bb*np.arctan(1/bb) + 4*logRmchb

    # Phi2
    term2 = 2*bb**2 * (4 - 4*bb*np.arctan(1/bb) - 3*np.log(1 + 1/bb**2))

    # adjustment
    if adjust is True:
        bb_too_large = bb > 1000
        term2[bb_too_large] = -10/3.

    Phi2 = (
        (4/3) - 2*np.log(1 + bb**2)
        + term2
        + 4*logRmchb
    )

    # adim
    f0 = 4*logRmchb + F2 - 4*fc*theta
    f1 = Phi1 - 4*logRmchb
    f2 = 0.5*(3*Phi1 - Phi2) - 4*logRmchb

    phi1 = f1 + f0
    phi2 = (4/3) * (1 - eps) * (f2 + f0)

    # cross-section in m2
    dcross_Ere_deps = (
        a0**2 * alpha**5 * atomic_nb * (atomic_nb + eta)
        * (phi1 * eps + phi2 / eps)
    )

    # change of variable to find derivative vs Eph (m2/eV)
    # deps / dEp = 1/(E_re + mc2)
    # m2 / eV
    dcross_Ere = dcross_Ere_deps / (E_re_eV + mc2_eV)

    # -------------
    # Optional anisotropy
    # -------------

    if costheta is not None:
        anis = anisotropy(
            costheta=costheta,
            gamma=gamma,
        )

        dcross_Ere = dcross_Ere * anis

    # -------------
    # format output
    # -------------

    dout = {
        'dcross_ei_Ere': {
            'data': dcross_Ere,
            'units': 'm2/eV',
        },
    }

    # -----------------
    # intermediates
    # -----------------

    if return_intermediates is True:

        dout.update({
            'RZ13a0': {
                'data': Rz3a0,
                'units': '',
            },
            'eta_inf': {
                'data': eta_inf,
                'units': '',
            },
        })

    return dout


def _check_dcross_ei(
    E_re_eV=None,
    E_ph_eV=None,
    atomic_nb=None,
    adjust=None,
    # optional anistropy
    costheta=None,
    # options
    return_intermediates=None,
):

    # -----------------
    # options
    # -----------------

    # adjust
    adjust = ds._generic_check._check_var(
        adjust, 'adjust',
        types=bool,
        default=True,
    )

    # return_intermediates
    return_intermediates = ds._generic_check._check_var(
        return_intermediates, 'return_intermediates',
        types=bool,
        default=False,
    )

    # -----------------
    # broadcastable
    # -----------------

    dout = {
        'E_re_eV': E_re_eV,
        'E_ph_eV': E_ph_eV,
        'atomic_nb': atomic_nb,
    }
    if costheta is not None:
        dout['costheta'] = costheta

    dout, shape = ds._generic_check._check_all_broadcastable(
        **dout,
    )

    lk = ['E_re_eV', 'E_ph_eV', 'atomic_nb']
    E_re_eV, E_ph_eV, atomic_nb = [dout[k0] for k0 in lk]
    if costheta is not None:
        costheta = dout['costheta']

    return (
        E_re_eV, E_ph_eV,
        atomic_nb, adjust,
        costheta,
        return_intermediates,
    )


# ##############################################################
# ##############################################################
#            plot Differencial Bremsstrahlung cross-section
# ##############################################################


def plot_dcross_ei(
    atomic_nb=None,
    Eplot_e=None,
    Eplot_ph=None,
    ang=None,
):

    # --------------
    # Check inputs
    # --------------

    # atomic number
    if atomic_nb is None:
        atomic_nb = 13

    # runaway
    if Eplot_e is None:
        Eplot_e = np.r_[1e6, 10e6, 15e6]
    E_e = Eplot_e

    # photon
    if Eplot_ph is None:
        Eplot_ph = np.r_[500e3, 5e6, 10e6]
    E_ph = np.linspace(Eplot_ph.min()*0.9, Eplot_ph.max()*1.1, 100)

    # angle
    if ang is None:
        ang = np.pi * np.linspace(0, 1, 50)

    # --------------
    # compute data
    # --------------

    # cross-section
    dcross = dcross_ei(
        E_re_eV=E_e[:, None, None],
        E_ph_eV=E_ph[None, :, None],
        atomic_nb=atomic_nb,
        adjust=True,
        # optional anistropy
        costheta=np.cos(ang[None, None, :]),
        # options
        return_intermediates=True,
    )['dcross_ei_Ere']['data']

    # --------------
    # prepare data
    # --------------

    # prepare
    indE_e = np.array([np.argmin(np.abs(E_e - ee)) for ee in Eplot_e])
    indE_ph = np.array([np.argmin(np.abs(E_ph - ee)) for ee in Eplot_ph])

    # str
    E_re_str = [
        f"{round(Eplot_e[ii]*1e-6, ndigits=1)} MeV"
        for ii in range(Eplot_e.size)
    ]
    E_ph_str = [
        f"{round(Eplot_ph[ii]*1e-6, ndigits=1)} MeV"
        for ii in range(Eplot_ph.size)
    ]

    # vmin, vmax
    vmin = np.min(np.log10(dcross[dcross > 0]))
    vmax = np.max(np.log10(dcross[dcross > 0]))
    dcross_plot = np.full(dcross.shape[1:], np.nan)

    # --------------
    # prepare figure
    # --------------

    tit = (
        "cross section vs E_e, E_ph, angle\n"
        f"Z = {atomic_nb}"
    )

    dmargin = {
        'left': 0.05, 'right': 0.95,
        'bottom': 0.05, 'top': 0.95,
        'wspace': 0.2, 'hspace': 0.2,
    }

    # --------------
    # figure
    # --------------

    # prepare figure
    fig = plt.figure(figsize=(14, 10))
    fig.suptitle(tit, size=12, fontweight='bold')
    gs = gridspec.GridSpec(ncols=3, nrows=indE_e.size, **dmargin)

    # --------------
    # plot
    # --------------

    axref = None
    axref1 = None
    dcol = {}
    for ii in range(indE_e.size):

        # ax2
        ax2 = fig.add_subplot(gs[ii, 0])
        ax2.set_xlabel(
            r'$log_{10}(E_{ph})$ (eV)',
            size=12,
            fontweight='bold',
        )
        ax2.set_ylabel('angle (rad)', size=12, fontweight='bold')

        # ax0 - anisotropy
        ax0 = fig.add_subplot(
            gs[ii, 1],
            aspect='equal',
            # adjustable='datalim',
            sharex=axref,
            sharey=axref,
        )
        ax0.set_xlabel(r'$\log_{10}(\sigma)$', size=12, fontweight='bold')
        ax0.set_ylabel(r'$\log_{10}(\sigma)$', size=12, fontweight='bold')
        ax0.set_title(f'E_re = {E_re_str[ii]}', size=12, fontweight='bold')

        # ax1
        ax1 = fig.add_subplot(gs[ii, 2], sharey=axref1)
        ax1.set_xlabel('E_ph (eV)', size=12, fontweight='bold')
        ax1.set_ylabel('cross-section (m2/eV)', size=12, fontweight='bold')

        # contour
        dcross_plot[...] = np.nan
        iok = np.isfinite(dcross[indE_e[ii], :, :])
        iok[iok] = dcross[indE_e[ii], iok] > 0
        dcross_plot[iok] = np.log10(dcross[indE_e[ii], iok])
        im = ax2.contourf(
            np.log10(E_ph),
            ang,
            dcross_plot.T,
            levels=15,
            vmin=vmin,
            vmax=vmax,
        )

        # contour
        for jj, iEph in enumerate(indE_ph):
            ll, = ax0.plot(
                (dcross_plot[iEph, :] - vmin) * np.cos(ang),
                (dcross_plot[iEph, :] - vmin) * np.sin(ang),
                '.-',
                label=f"E_ph = {E_ph_str[jj]}",
                c=dcol.get(jj),
            )
            if ii == 0:
                dcol[jj] = ll.get_color()

            ax2.axvline(np.log10(E_ph[iEph]), c=dcol[jj], ls='--', lw=1)

        ax0.legend()

        # slice
        ax1.loglog(
            E_ph,
            dcross_plot[:, 0],
            '.-',
            c='k',
            label=f"{ang[0]*180/np.pi:.1f} deg",
        )
        ax1.loglog(
            E_ph,
            dcross_plot[:, -1],
            '.--',
            c='k',
            label=f"{ang[-1]*180/np.pi:.1f} deg",
        )
        ax1.legend()

        plt.colorbar(im, ax=ax2, label=r'$\log_{10}(\sigma)$')

        ax0.axhline(0, c='k', lw=1, ls='--')
        ax0.axvline(0, c='k', lw=1, ls='--')
        ax0.set_ylim(bottom=0)

        if ii == 0:
            axref = ax0
            axref1 = ax1

    return {'map': ax2, 'anisotropy': ax0, 'curve': ax1}


# #################################################################
# #################################################################
#                Plot vs SALVAT
# #################################################################


def plot_dcross_vs_Salvat(
):

    # --------------
    # input data
    # --------------

    E_re = np.r_[1, 10, 100]*1e6
    E_ph = np.linspace(0.01*np.min(E_re), np.max(E_re), 10000)

    E_re_str = [f"{round(ee*1e-6)} MeV" for ee in E_re]

    # --------------
    # prepare figure
    # --------------

    # cross-section
    dcross = dcross_ei(
        E_re_eV=E_re[None, :],
        E_ph_eV=E_ph[:, None],
        atomic_nb=13,
        adjust=True,
        # optional anistropy
        costheta=None,
        # options
        return_intermediates=True,
    )['dcross_ei_Ere']['data']

    # beta
    beta = _utils.convert_momentum_velocity_energy(
        energy_kinetic_eV=E_re,
    )['beta']['data']

    # --------------
    # prepare figure
    # --------------

    ERE = np.r_[1., 10., 100.]*1e6
    RSB = [
        0., 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 0.95, 0.97, 0.99, 1.,
    ]
    ySB = np.array([
        [14.03, 10.17, 7.878, 6.296, 5.092, 4.111, 3.287, 2.564, 1.913, 1.312,
         1.001, 0.87, 0.735, 0.665],
        [16.45, 12.68, 10.80, 9.34, 8.077, 7.057, 6.165, 5.309, 4.388, 3.105,
         2.111, 1.588, 0.945, 0.52],
        [16.36, 13.24, 11.98, 10.94, 10.10, 9.454, 8.967, 8.585, 8.173, 7.262,
         6.022, 5.004, 2.964, 0.496],
    ])

    # --------------
    # prepare figure
    # --------------

    fig = plt.figure(figsize=(14, 8))
    fig.suptitle(
        "Comparing to literature for Al (Z=13)\n"
        "Salvat et al. 1992",
        size=14,
        fontweight='bold',
    )

    ylab = r'$\left(\frac{\beta}{Z}\right)^2 E_{ph} \frac{d\sigma}{dE_{ph}}$'

    # --------------
    # axes
    # --------------

    dax = {}
    ax0 = None
    for ii, ee in enumerate(E_re):
        ax = fig.add_axes(
            [0.08 + ii*(0.27 + 0.05), 0.08, 0.27, 0.75],
            sharex=ax0,
            sharey=ax0,
        )

        ax.set_xlabel(r"$\frac{E_{ph}}{E_{re}}$", size=12, fontweight='bold')
        ax.set_title(
            r"$E_{re}$" + f" = {E_re_str[ii]}",
            fontweight='bold',
            size=12,
        )

        if ii == 0:
            ax0 = ax
            ax.set_ylabel(
                ylab + ' (mb)',
                size=12,
                fontweight='bold',
            )

        dax[ii] = ax

    # --------------
    # plot
    # --------------

    for ii, ee in enumerate(E_re):
        ax = dax[ii]

        indE = np.argmin(np.abs(E_re - ee))
        indES = np.argmin(np.abs(ERE - ee))

        ax.plot(
            E_ph / E_re[indE],
            (beta[ii] / 13**2) * E_ph * dcross[:, indE] * 1e31,
            '.-',
            label=f'calc (E_re = {E_re_str[ii]})',
        )

        ax.plot(
            RSB,
            ySB[indES, :],
            'ok',
            markerfacecolor='None',
            label=f'ref (E_re = {ee*1e-6} MeV)',
        )
        ax.legend()

    ax0.set_xlim(0, 1)

    return dax





import numpy as np
import scipy.constants as scpct
import datastock as ds


# ####################################################
# ####################################################
#        Maxwellian-averaged
# ####################################################


def get_bremss_maxwell(
    # parameters
    kTe_eV=None,
    ne_m3=None,
    Zeff=None,
    # expressed on
    E_ph_eV=None,
    lambda_m=None,
):

    # -------------
    # check inputs
    # -------------

    kTe_eV, ne_m3, Zeff, E_ph_eV = _check(
        # parameters
        kTe_eV=kTe_eV,
        ne_m3=ne_m3,
        Zeff=Zeff,
        # expressed on
        E_ph_eV=E_ph_eV,
        lambda_m=lambda_m,
    )

    # -------------
    # Prepare
    # -------------

    # e, C
    e = scpct.e

    # h, eV.s
    h_eVs = scpct.h / e

    # hbar, eV.s
    hbar_eVs = h_eVs / (2.*np.pi)

    # fine structure
    alpha = scpct.alpha

    # mec2, eV
    mec2_eV = scpct.m_e * scpct.c**2 / e

    # mec eV.s/m
    mec_eVsm = mec2_eV / scpct.c

    # first Borh radius, m
    a0 = hbar_eVs / (mec_eVsm * alpha)

    # ER, eV
    ER_eV = hbar_eVs * scpct.c * alpha / (2 * a0)

    # gaunt
    gaunt = 1.

    # -------------
    # compute
    # -------------

    # Taken from
    # [1] Vezinet, PhD thesis, 2013, p. 51, eq. 41

    # coef0, m^3
    coef0_m3 = 2**5 * np.sqrt(np.pi) * (alpha * a0)**3 / (3*np.sqrt(3))

    # kk, ph.m3/s/sr
    kk = (
        coef0_m3
        * (ER_eV / h_eVs)
        * np.sqrt(ER_eV / kTe_eV)
        * np.exp(-E_ph_eV / kTe_eV)
        * gaunt
    )

    # emiss, ph/m3/s
    emiss = ne_m3**2 * Zeff * kk

    # -------------
    # outputs
    # -------------

    dout = {
        'emiss': {
            'data': emiss,
            'units': 'ph/m3/s/sr',
        },
    }

    return dout


def _check(
    # parameters
    kTe_eV=None,
    ne_m3=None,
    Zeff=None,
    # expressed on
    E_ph_eV=None,
    lambda_m=None,
):

    # -------------
    # E vs lambda
    # -------------

    if E_ph_eV is None:

        if lambda_m is None:
            msg = 'Please provide either E_ph_eV xor lambda_m!'
            raise Exception(msg)

        # energy eV
        E_ph_eV = scpct.h * scpct.c / lambda_m / scpct.e

    # -------------
    # broadcast
    # -------------

    dparams, shape = ds._generic_check._check_all_broadcastable(
        # parameters
        kTe_eV=kTe_eV,
        ne_m3=ne_m3,
        Zeff=Zeff,
        # expressed on
        E_ph_eV=E_ph_eV,
    )

    lk = ['kTe_eV', 'ne_m3', 'Zeff', 'E_ph_eV']
    return [dparams[k0] for k0 in lk]


# ####################################################
# ####################################################
#        Cross-section
# ####################################################


def get_dcross_ei():
    """ Return a differential cross-section for thin-target bremsstrahlung

    Uses the BHE or BE formulas [1]:
        - BHE: Bethe-Heitler-Elwert [2]
        - BE: Elwert-Haug [3]

    Uses:
        [1] Y. Peysson, "Rayonnement electromagnetique des plasmas"
        [2]
        [3]

    """

    # -------------
    # check input
    # -------------

    # -------------
    # prepare
    # -------------

    # -------------
    # compute
    # -------------

    # -------------
    # format output
    # -------------

    return dout



# -*- coding: utf-8 -*-
"""
Created on Wed Jun 12 18:23:48 2024

@author: dvezinet
"""


import numpy as np
import datastock as ds


from . import _class2_check as _check


# ###############################################################
# ###############################################################
#                  tangency radius
# ###############################################################


def _tangency_radius(
    coll=None,
    key=None,
    key_cam=None,
    quantity=None,
    # limits
    segment=None,
    lim_to_segments=None,
    # tangency radius
    axis_pt=None,
    axis_vect=None,
):

    # --------------
    # check inputs
    # --------------

    (
        key, quantity, segment, lim_to_segments, axis_pt, axis_vect,
    ) = _tangency_radius_check(
        coll=coll,
        key=key,
        key_cam=key_cam,
        quantity=quantity,
        segment=segment,
        lim_to_segments=lim_to_segments,
        axis_pt=axis_pt,
        axis_vect=axis_vect,
    )

    # -----------
    # compute
    # -----------

    # --------
    # prepare

    (
     pts_x, pts_y, pts_z,
     ABx, ABy, ABz,
     AOx, AOy, AOz,
     ABvn2, AOvn2, B,
     i0,
     ) = _tangency_radius_prepare(
         coll=coll,
         key=key,
         quantity=quantity,
         segment=segment,
         axis_pt=axis_pt,
         axis_vect=axis_vect,
    )

    # --------
    # radius

    if quantity == 'tangency_radius':
        kk = np.zeros(ABx.shape, dtype=float)
        iok = ABvn2 > 0.
        kk[iok] = -B[iok] / ABvn2[iok]

        # ------------------------
        # lim_to_segments

        if lim_to_segments is True:
            i0 = kk < 0.
            i1 = kk > 1.

            kk[i0] = 0.
            kk[i1] = 1.

        out = np.sqrt(kk**2 * ABvn2 + 2*kk*B + AOvn2)

    elif quantity == 'length':
        out = np.sqrt(ABx**2 + ABy**2 + ABz**2)
        kk = None

    else:
        kdata = coll.dobj['rays'][key]['alpha']
        out = coll.ddata[kdata]['data']
        kk = None

    # -------
    # ref

    if segment is None:
        ref = coll.dobj['rays'][key]['ref']
    elif segment.size == 1:
        ref = coll.dobj['rays'][key]['ref'][1:]
        out = out[0]
        if quantity == 'tangency radius':
            kk = kk[0]
    else:
        ref = None

    # -------
    # return

    return out, kk, ref


# ###############################################################
#              tangency radius - check
# ###############################################################


def _tangency_radius_check(
    coll=None,
    key=None,
    key_cam=None,
    quantity=None,
    segment=None,
    lim_to_segments=None,
    axis_pt=None,
    axis_vect=None,
):

    # key
    key = _check._check_key(coll=coll, key=key, key_cam=key_cam)

    # quantity
    quantity = ds._generic_check._check_var(
        quantity, 'quantity',
        types=str,
        default='tangency_radius',
        allowed=['alpha', 'tangency_radius', 'length'],
    )

    # lim_to_segments
    lim_to_segments = ds._generic_check._check_var(
        lim_to_segments, 'lim_to_segments',
        types=bool,
        default=False,
    )

    # segment
    if segment is not None:
        segment = np.atleast_1d(segment).astype(int).ravel()

    # --------
    # tangency radius-specific

    if quantity == 'tangency_radius':
        # axis_pt
        if axis_pt is None:
            axis_pt = [0., 0., 0.]

        axis_pt = ds._generic_check._check_flat1darray(
            axis_pt, 'axis_pt',
            dtype=float,
            size=3,
            can_be_None=False,
        )

        # axis_vect
        if axis_vect is None:
            axis_vect = [0., 0., 1.]

        axis_vect = ds._generic_check._check_flat1darray(
            axis_vect, 'axis_vect',
            dtype=float,
            size=3,
            norm=True,
            can_be_None=False,
        )

    return key, quantity, segment, lim_to_segments, axis_pt, axis_vect


# ###############################################################
#            tangency radius - prepare data
# ###############################################################


def _tangency_radius_prepare(
    coll=None,
    key=None,
    quantity=None,
    segment=None,
    axis_pt=None,
    axis_vect=None,
):

    pts_x, pts_y, pts_z = coll.get_rays_pts(key=key)
    i0 = np.arange(0, pts_x.shape[0])

    # select segment
    if segment is not None:
        npts = pts_x.shape[0]
        segment[segment < 0] = npts - 1 + segment[segment < 0]

        iseg = np.r_[segment, segment[-1] + 1]
        i0 = i0[iseg]
        pts_x = pts_x[iseg, :]
        pts_y = pts_y[iseg, :]
        pts_z = pts_z[iseg, :]


    # define vectors
    ABx = np.diff(pts_x, axis=0)
    ABy = np.diff(pts_y, axis=0)
    ABz = np.diff(pts_z, axis=0)

    if quantity == 'tangency_radius':
        AOx = axis_pt[0] - pts_x[:-1, ...]
        AOy = axis_pt[1] - pts_y[:-1, ...]
        AOz = axis_pt[2] - pts_z[:-1, ...]

        # --------
        # get kk

        # Eq:
        # k^2 |AB x v|^2 + 2k((AB.v)(AO.v) - AB.AO) + |AO x v|^2 = dist^2
        #
        # minimum:
        # k |AB x v|^2 + (AB.v)(AO.v) - AB.AO = 0

        ABvn2 = (
            (ABy*axis_vect[2] - ABz*axis_vect[1])**2
            + (ABz*axis_vect[0] - ABx*axis_vect[2])**2
            + (ABx*axis_vect[1] - ABy*axis_vect[0])**2
            )
        AOvn2 = (
            (AOy*axis_vect[2] - AOz*axis_vect[1])**2
            + (AOz*axis_vect[0] - AOx*axis_vect[2])**2
            + (AOx*axis_vect[1] - AOy*axis_vect[0])**2
            )

        ABv = ABx*axis_vect[0] + ABy*axis_vect[1] + ABz*axis_vect[2]
        AOv = AOx*axis_vect[0] + AOy*axis_vect[1] + AOz*axis_vect[2]
        ABAO = ABx*AOx + ABy*AOy + ABz*AOz
        B = ABv*AOv - ABAO

    else:
        AOx, AOy, AOz = None, None, None
        ABvn2, AOvn2, B = None, None, None
        i0 = None

    return (
        pts_x, pts_y, pts_z,
        ABx, ABy, ABz,
        AOx, AOy, AOz,
        ABvn2, AOvn2, B,
        i0,
    )


# ###############################################################
# ###############################################################
#                  intersect radius
# ###############################################################


def intersect_radius(
    coll=None,
    key=None,
    key_cam=None,
    axis_pt=None,
    axis_vect=None,
    axis_radius=None,
    segment=None,
    lim_to_segments=None,
    return_pts=None,
    return_itot=None,
):

    # --------------
    # check inputs
    # --------------

    (
     key, quantity, segment, lim_to_segments, axis_pt, axis_vect,
     ) = _tangency_radius_check(
         coll=coll,
         key=key,
         key_cam=key_cam,
         quantity='tangency_radius',
         axis_pt=axis_pt,
         axis_vect=axis_vect,
         segment=segment,
         lim_to_segments=lim_to_segments,
    )

    # axis_radius
    axis_radius = ds._generic_check._check_var(
        axis_radius, 'axis_radius',
        types=(float, int),
        sign='> 0.',
    )

    # return_pts
    return_pts = ds._generic_check._check_var(
        return_pts, 'return_pts',
        types=bool,
        default=False,
    )

    if return_pts is True and lim_to_segments is False:
        msg = (
            "return_pts requires lim_to_segments = True"
            )
        raise Exception(msg)

    # return_itot
    return_itot = ds._generic_check._check_var(
        return_itot, 'return_itot',
        types=bool,
        default=False,
    )

    # -----------
    # compute
    # -----------

    # --------
    # prepare

    (
        pts_x, pts_y, pts_z,
        ABx, ABy, ABz,
        AOx, AOy, AOz,
        ABvn2, AOvn2, B,
        i0,
    ) = _tangency_radius_prepare(
         coll=coll,
         key=key,
         quantity=quantity,
         segment=segment,
         axis_pt=axis_pt,
         axis_vect=axis_vect,
    )

    # pre-select according to rad_min
    kmin = np.zeros(ABx.shape, dtype=float)

    iok = ABvn2 > 0.
    kmin[iok] = -B[iok] / ABvn2[iok]
    rad_min2 = kmin**2 * ABvn2 + 2*kmin*B + AOvn2
    rad_min2[np.abs(rad_min2) < 1e-9] = 0.
    rad_min = np.sqrt(rad_min2)

    # prepare solutions
    iin = (rad_min < axis_radius) & iok

    # there can be up to 2 solutions
    k0 = np.full(ABx.shape, np.nan)
    k1 = np.full(ABx.shape, np.nan)

    if np.any(iin):
        delta = B[iin]**2 - ABvn2[iin]*(AOvn2[iin] - axis_radius**2)

        k0[iin] = (-B[iin] - np.sqrt(delta)) / ABvn2[iin]
        k1[iin] = (-B[iin] + np.sqrt(delta)) / ABvn2[iin]

    # ----------------
    # lim_to_segments

    if lim_to_segments is True:

        ind = np.copy(iin)
        ind[ind] = (k0[ind] < 0) & (k1[ind] > 0)
        k0[ind] = 0.

        ind = np.copy(iin)
        ind[ind] = (k0[ind] < 1) & (k1[ind] > 1)
        k1[ind] = 1.

        ind = np.copy(iin)
        ind[ind] = (k0[ind] >= 0.) & (k0[ind] <= 1.)
        k0[~ind] = np.nan

        ind = np.copy(iin)
        ind[ind] = (k1[ind] >= 0.) & (k1[ind] <= 1.)
        k1[~ind] = np.nan

    # ----------------------
    # additional derivations

    iok = np.isfinite(k0) & np.isfinite(k1)

    if return_pts is True:

        iok2 = np.copy(iok)
        iok2[iok] = k1[iok] < 1.

        false = np.zeros(tuple(np.r_[1, iok.shape[1:]]), dtype=bool)
        iok0 = np.concatenate((iok, false), axis=0)
        iok02 = np.concatenate((iok2, false), axis=0)
        iok12 = np.concatenate((false, iok2), axis=0)

        # Make sure there a single continued sequence per ray
        # build index and check continuity

        px = np.full(pts_x.shape, np.nan)
        py = np.full(pts_x.shape, np.nan)
        pz = np.full(pts_x.shape, np.nan)

        px[iok0] = pts_x[iok0] + k0[iok] * ABx[iok]
        py[iok0] = pts_y[iok0] + k0[iok] * ABy[iok]
        pz[iok0] = pts_z[iok0] + k0[iok] * ABz[iok]

        px[iok12] = pts_x[iok02] + k1[iok2] * ABx[iok2]
        py[iok12] = pts_y[iok02] + k1[iok2] * ABy[iok2]
        pz[iok12] = pts_z[iok02] + k1[iok2] * ABz[iok2]
        px[-1, ...] = pts_x[-2, ...] + k1[-1, ...] * ABx[-1, ...]
        py[-1, ...] = pts_y[-2, ...] + k1[-1, ...] * ABy[-1, ...]
        pz[-1, ...] = pts_z[-2, ...] + k1[-1, ...] * ABz[-1, ...]

    if return_itot is True:

        i02 = np.full(pts_x.shape, np.nan)
        for ii in range(pts_x.shape[0]):
            i02[ii, ...] = i0[ii]

        itot = np.full(pts_x.shape, np.nan)

        itot[iok0] = i02[iok0] + k0[iok]
        itot[iok12] = i02[iok02] + k1[iok2]
        itot[-1, ...] = i02[-2, ...] + k1[-1, ...]

    # ------
    # return

    if return_pts is True and return_itot is True:
        return k0, k1, iok, px, py, pz, itot
    elif return_pts is True:
        return k0, k1, iok, px, py, pz
    elif return_itot is True:
        return k0, k1, iok, itot
    else:
        return k0, k1, iok



# -*- coding: utf-8 -*-
""" Basic tools for formatting 2d polygons

"""


import numpy as np
import datastock as ds


# ###########################################################
# ###########################################################
#              check
# ###########################################################


def check(
    x0=None,
    x1=None,
    key=None,
    # options
    closed=None,
    clockwise=None,
):
    """ High-level routine to format a 2d polygon


    Parameters
    ----------
    x0 : sequence
        1st coordinate of the polygon
    x1 : sequence
        2nd coordinate of the polygon
    key : str / None
        To inform error messages
    closed : bool / None
        whether the polygon should be closed
    clockwise : bool / None
        whether the polygon should be clockwise

    Returns
    -------
    dout : dict
        {
            'x0': np.ndarray,
            'x0': np.ndarray,
            'closed': bool,
            'clockwise': bool,
            'key': str,
        }

    """

    # -------------
    # check inputs
    # -------------

    key, close, clockwise = _check(
        key=key,
        closed=closed,
        clockwise=clockwise,
    )

    # -------------
    # check x0, x1
    # -------------

    x0 = ds._generic_check._check_flat1darray(
        x0, 'x0',
        dtype=float,
        can_be_None=False,
        extra_msg=f"x0 of polygon '{key}'",
    )

    x1 = ds._generic_check._check_flat1darray(
        x1, 'x1',
        dtype=float,
        can_be_None=False,
        size=x0.size,
        extra_msg=f"x1 of polygon '{key}'",
    )

    # -------------
    # closed
    # -------------

    is_closed = np.allclose(np.r_[x0[0], x1[0]], np.r_[x0[-1], x1[-1]])
    if is_closed is True:
        x0_closed = x0
        x1_closed = x1

    else:
        ind_closed = np.r_[np.arange(0, x0.size), 0]
        x0_closed = x0[ind_closed]
        x1_closed = x1[ind_closed]

    # -------------
    # no duplicates
    # -------------

    uni = np.unique([x0_closed[:-1], x1_closed[:-1]], axis=1)
    if uni.shape[1] < (x0_closed.size - 1):
        ndup = x0.size - 1 - uni.shape[1]
        msg = (
            f"Polygon 2d '{key}' seems to have {ndup} duplicate points!\n"
            "\t- x0 = {x0_closed[:-1]}\n"
            "\t- x1 = {x1_closed[:-1]}\n"
        )
        raise Exception(msg)

    # -------------
    # clockwise
    # -------------

    # is already ?
    is_cw = is_clockwise(x0_closed, x1_closed)

    # adjust
    if is_cw != clockwise:
        x0_closed, x1_closed = x0_closed[::-1], x1_closed[::-1]

    # -------------
    # return
    # -------------

    dout = {
        'x0': x0_closed if closed else x0_closed[:-1],
        'x1': x1_closed if closed else x1_closed[:-1],
        'key': key,
        'closed': closed,
        'clockwise': clockwise,
    }

    return dout

# ###########################################################
# ###########################################################
#              check
# ###########################################################


def _check(
    key=None,
    closed=None,
    clockwise=None,
):

    # ---------------
    # key
    # ---------------

    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        default='',
    )

    # ---------------
    # closed
    # ---------------

    closed = ds._generic_check._check_var(
        closed, 'closed',
        types=bool,
        default=False,
    )

    # ---------------
    # clockwise
    # ---------------

    clockwise = ds._generic_check._check_var(
        clockwise, 'clockwise',
        types=bool,
        default=True,
    )

    return key, closed, clockwise


# ###########################################################
# ###########################################################
#              is clockwise
# ###########################################################


def is_clockwise(x0, x1):
    area_signed = np.sum((x0[1:] - x0[:-1]) * (x1[1:] + x1[:-1]))
    return area_signed > 0




# #############################################################################
# #############################################################################
#           data of mobile based on indices
# #############################################################################


def _set_dbck(lax=None, daxes=None, dcanvas=None, dmobile=None):
    """ Update background of relevant axes (ex: in case of resizing) """

    # Make all invisible
    for k0 in lax:
        for k1 in daxes[k0]['mobile']:
            dmobile[k1]['handle'].set_visible(False)

    # Draw and reset bck
    lcan = set([daxes[k0]['canvas'] for k0 in lax])
    for k0 in lcan:
        dcanvas[k0]['handle'].draw()

    # set bck (= bbox copy)
    for k0 in lax:
        # ax.draw(self.can.renderer)
        daxes[k0]['bck'] = dcanvas[
            daxes[k0]['canvas']
        ]['handle'].copy_from_bbox(daxes[k0]['handle'].bbox)

    # Redraw
    for k0 in lax:
        for k1 in daxes[k0]['mobile']:
            dmobile[k1]['handle'].set_visible(dmobile[k1]['visible'])
            # ax.draw(self.can.renderer)

    for k0 in lcan:
        dcanvas[k0]['handle'].draw()


# #############################################################################
# #############################################################################
#           Update number of visible indices
# #############################################################################


def _get_nn_ii_group(
    nmax=None,
    nmaxcur=None,
    indcur=None,
    ctrl=None,
    shift=None,
    group=None,
):
    """"""

    if shift and nmaxcur == nmax:
        msg = f"Max nb. of plots reached for group '{group}': {nmax}"
        print(msg)
        return False

    if ctrl:
        nn = 0
        ii = 0
    elif shift:
        nn = int(nmaxcur) + 1
        ii = nn - 1
    else:
        nn = int(nmaxcur)
        ii = int(indcur)
    return nn, ii


def _update_indices_nb(group=None, dgroup=None, ctrl=None, shift=None):
    """"""
    out = _get_nn_ii_group(
        nmax=dgroup[group]['nmax'],
        nmaxcur=dgroup[group]['nmaxcur'],
        indcur=dgroup[group]['indcur'],
        ctrl=ctrl,
        shift=shift,
        group=group,
    )
    if out is False:
        return False
    else:
        dgroup[group]['nmaxcur'] = out[0]
        dgroup[group]['indcur'] = out[1]


# #############################################################################
# #############################################################################
#           data of mobile based on indices
# #############################################################################


def get_fupdate(handle=None, dtype=None, norm=None, bstr=None):
    if dtype == 'xdata':
        func = lambda val, handle=handle: handle.set_xdata(val)
    elif dtype == 'ydata':
        func = lambda val, handle=handle: handle.set_ydata(val)
    elif dtype in ['data']:   # Also works for imshow
        func = lambda val, handle=handle: handle.set_data(val)
    elif dtype in ['alpha']:   # Also works for imshow
        func = lambda val, handle=handle, norm=norm: handle.set_alpha(norm(val))
    elif dtype == 'txt':
        func = lambda val, handle=handle, bstr=bstr: handle.set_text(bstr.format(val))
    return func


def _update_mobile_data(
    func=None,
    kref=None,
    kdata=None,
    iref=None,
    ddata=None,
):
    """"""

    # if handle.__class__.__name__ == 'Line2D':

    if kdata == 'index':
        func(iref)

    elif ddata[kdata]['data'].ndim == 1:
        func(ddata[kdata]['data'][iref])

    elif ddata[kdata]['data'].ndim == 2:

        idim = ddata[kdata]['ref'].index(kref)
        if idim == 0:
            func(ddata[kdata]['data'][iref, :])
        else:
            func(ddata[kdata]['data'][:, iref])

    elif ddata[kdata]['data'].ndim == 3:

        idim = ddata[kdata]['ref'].index(kref)
        if idim == 0:
            func(ddata[kdata]['data'][iref, :, :])
        elif idim == 1:
            func(ddata[kdata]['data'][:, iref, :])
        elif idim == 2:
            func(ddata[kdata]['data'][:, :, iref])


def _update_mobile(k0=None, dmobile=None, dref=None, ddata=None):
    """ Update mobile objects data """

    func = dmobile[k0]['func']
    dtype = dmobile[k0]['dtype']

    kref = dmobile[k0]['ref']
    kdata = dmobile[k0]['data']
    iref = [dref[rr]['indices'][dmobile[k0]['ind']] for rr in kref]

    if kref[0] is not None:
        _update_mobile_data(
            func=func,
            kref=kref[0],
            kdata=kdata[0],
            iref=iref[0],
            ddata=ddata,
        )

    if len(kref) > 1 and kref[1] is not None:
        _update_mobile_data(
            func=func,
            kref=kref[1],
            kdata=kdata[1],
            iref=iref[1],
            ddata=ddata,
        )



# -*- coding: utf-8 -*-


import warnings


import numpy as np


import datastock as ds


# ##############################################################
# ##############################################################
#           Local to global coordinates
# ##############################################################


def _get_x01toxyz(
    coll=None,
    key=None,
    asplane=None,
):

    # ---------
    # key

    key, cls = coll.get_optics_cls(optics=key)
    key, cls = key[0], cls[0]
    dgeom = coll.dobj[cls][key]['dgeom']

    # asplane
    asplane = ds._generic_check._check_var(
        asplane, 'asplane',
        types=bool,
        default=False,
    )

    # -------------------
    #     Planar
    # -------------------

    if dgeom['type'] == 'planar' or asplane is True:

        def x01toxyz(
            x0=None,
            x1=None,
            # surface
            cent=dgeom['cent'],
            e0=dgeom['e0'],
            e1=dgeom['e1'],
        ):
            """ Coordinate transform """

            return (
                cent[0] + x0*e0[0] + x1*e1[0],
                cent[1] + x0*e0[1] + x1*e1[1],
                cent[2] + x0*e0[2] + x1*e1[2],
            )

    # ----------------
    #   Cylindrical
    # ----------------

    elif dgeom['type'] == 'cylindrical':

        iplan = np.isinf(dgeom['curve_r']).nonzero()[0][0]
        eax = ['e0', 'e1'][iplan]
        erot = ['e0', 'e1'][1-iplan]

        rc = dgeom['curve_r'][1 - iplan]
        rcs = np.sign(rc)
        rca = np.abs(rc)

        def x01toxyz(
            x0=None,
            x1=None,
            # surface
            O=dgeom['cent'] + dgeom['nin'] * rc,
            rcs=rcs,
            rca=rca,
            eax=dgeom[eax],
            erot=dgeom[erot],
            # local coordinates
            nin=dgeom['nin'],
            iplan=iplan,
        ):
            """ Coordinate transform """

            if iplan == 0:
                xx, theta = x0, x1
            else:
                xx, theta = x1, x0

            nox = np.cos(theta)*(-rcs*nin[0]) + np.sin(theta)*erot[0]
            noy = np.cos(theta)*(-rcs*nin[1]) + np.sin(theta)*erot[1]
            noz = np.cos(theta)*(-rcs*nin[2]) + np.sin(theta)*erot[2]

            return (
                O[0] + xx*eax[0] + rca*nox,
                O[1] + xx*eax[1] + rca*noy,
                O[2] + xx*eax[2] + rca*noz,
            )

    # ----------------
    #   Spherical
    # ----------------

    elif dgeom['type'] == 'spherical':

        rc = dgeom['curve_r'][0]
        rcs = np.sign(rc)
        rca = np.abs(rc)

        def x01toxyz(
            x0=None,
            x1=None,
            # surface
            O=dgeom['cent'] + dgeom['nin'] * rc,
            rcs=rcs,
            rca=rca,
            # local coordinates
            nin=dgeom['nin'],
            e0=dgeom['e0'],
            e1=dgeom['e1'],
        ):

            dtheta, phi = x1, x0

            ephix = np.cos(phi)*(-rcs*nin[0]) + np.sin(phi)*e0[0]
            ephiy = np.cos(phi)*(-rcs*nin[1]) + np.sin(phi)*e0[1]
            ephiz = np.cos(phi)*(-rcs*nin[2]) + np.sin(phi)*e0[2]

            nox = np.cos(dtheta)*ephix + np.sin(dtheta)*e1[0]
            noy = np.cos(dtheta)*ephiy + np.sin(dtheta)*e1[1]
            noz = np.cos(dtheta)*ephiz + np.sin(dtheta)*e1[2]

            return (
                O[0] + rca * nox,
                O[1] + rca * noy,
                O[2] + rca * noz,
            )

    # ----------------
    #   Toroidal
    # ----------------

    elif dgeom['type'] == 'toroidal':

        raise NotImplementedError()

    return x01toxyz





from ._xray_thick_target import anisotropy as get_xray_thick_anisotropy
from ._xray_thick_target import dcross_ei as get_xray_thick_dcross_ei
from ._xray_thick_target import plot_dcross_ei as plot_xray_thick_dcross_ei
from ._xray_thick_target import plot_dcross_vs_Salvat as plot_xray_thick_dcross_ei_vs_Salvat

from ._xray_thin_target import get_bremss_maxwell as get_xray_thin_bremss_maxwell






import numpy as np
import scipy.constants as scpct



def compute_bremzeff(Te=None, ne=None, zeff=None, lamb=None):
    """ Return the bremsstrahlun spectral radiance at lamb

    The plasma conditions are set by:
        - Te   (eV)
        - ne   (/m3)
        - zeff (adim.)

    The wavelength is set by the diagnostics
        - lamb (m)

    The vol. spectral emis. is returned in ph / (s.m3.sr.m)

    The computation requires an intermediate : gff(Te, zeff)
    """
    ktkeV = Te * 1.e-3
    ktJ = Te * scpct.e
    gff = 5.54 - (3.11-np.log(ktkeV))*(0.69-0.13/zeff)
    Const = ((scpct.e**6/(scpct.h*scpct.c**3*(np.pi*scpct.epsilon_0)**3))
             * np.sqrt(np.pi/(864.*scpct.m_e**3)))
    hc = scpct.h*scpct.c
    emis = Const/lamb * ne**2*zeff * np.exp(-hc/(lamb*ktJ)) * gff/np.sqrt(ktJ)
    units = r'ph / (s.m3.sr.m)'
    return emis, units


def compute_fangle(BR=None, BPhi=None, BZ=None, ne=None, lamb=None):
    """ The the vector quantity to be integrated on LOS to get faraday angle

    fangle = int_LOS ( abs(sca(quant, u_LOS)) )
    Where:
        quant = C * lamb**2 * ne * Bv

    With:
        - C = 2.615e-13  (1/T)
        - ne    (/m3)
        - lamb  (m)
        - Bv    (T)

    The resulting faraday angle (after integration) will be in radians
    """
    const = scpct.e**3 / (8.*scpct.pi**2
                          * scpct.epsilon_0 * scpct.m_e**2 * scpct.c**3)
    quant = const * lamb**2 * ne * np.array([BR, BPhi, BZ])
    units = r'rad / m'
    return quant, units



# -*- coding: utf-8 -*-
"""
Created on Wed Feb 21 14:32:58 2024

@author: dvezinet
"""


import os
import itertools as itt


import numpy as np
import datastock as ds


from . import _class02_save2stp


# #################################################################
# #################################################################
#          Default values
# #################################################################


_NAME = 'test'


# #################################################################
# #################################################################
#          Main
# #################################################################


def main(
    # ---------------
    # input from tofu
    coll=None,
    key=None,
    key_cam=None,
    key_optics=None,
    # ---------------
    # options
    factor=None,
    color=None,
    empty_name=None,
    # ---------------
    # saving
    pfe_save=None,
    overwrite=None,
):


    # ----------------
    # check inputs
    # --------------

    (
        key_optics,
        factor, color,
        iso,
        pfe_save, overwrite,
    ) = _check(
        coll=coll,
        key=key,
        key_cam=key_cam,
        key_optics=key_optics,
        # options
        factor=factor,
        color=color,
        # saving
        pfe_save=pfe_save,
        overwrite=overwrite,
    )

    fname = os.path.split(pfe_save)[-1][:-4]

    # -------------
    # extract and pre-format data
    # -------------

    dptsx, dptsy, dptsz, duvect, dcor_ptsvect = _extract(
        coll=coll,
        key_optics=key_optics,
    )

    # scaling factor
    for k0 in dptsx.keys():
        dptsx[k0] = factor * dptsx[k0]
        dptsy[k0] = factor * dptsy[k0]
        dptsz[k0] = factor * dptsz[k0]

    # ---------------
    # get color dict
    # ---------------

    dcolor = _class02_save2stp._get_dcolor(dptsx=dptsx, color=color)

    # ----------------
    # get file content
    # ----------------

    # HEADER
    msg_header = _class02_save2stp._get_header(
        fname=fname,
        iso=iso,
    )

    # DATA
    msg_data = _get_data(
        dptsx=dptsx,
        dptsy=dptsy,
        dptsz=dptsz,
        duvect=duvect,
        dcor_ptsvect=dcor_ptsvect,
        fname=fname,
        # options
        dcolor=dcolor,
        # norm
        iso=iso,
        # empty_name
        empty_name=empty_name,
    )

    # -------------
    # save to stp
    # -------------

    _class02_save2stp._save(
        msg=msg_header + "\n" + msg_data,
        pfe_save=pfe_save,
        overwrite=overwrite,
    )

    return


# #################################################################
# #################################################################
#          check
# #################################################################


def _check(
    coll=None,
    key=None,
    key_cam=None,
    key_optics=None,
    # options
    outline_only=None,
    factor=None,
    color=None,
    # saving
    pfe_save=None,
    overwrite=None,
    ext='stp',
):

    # ---------------
    # diag vs optics
    # ---------------

    lc = [
        key is not None and key_optics is None,
        key is None and key_cam is None and key_optics is not None,
    ]

    if np.sum(lc) != 1:
        msg = (
            "Please provide either (key, key_cam) xor key_optics:\n"
            "\t- (key, key_cam): to save all optcis associated to a diag/cam/n"
            "\t- key_optics: a list of arbitrary optics (apert, cryst, cam...)"
            "\nProvided:\n"
            f"\t- key: {key}\n"
            f"\t- key_cam: {key_cam}\n"
            f"\t- key_optics: {key_optics}\n"
        )
        raise Exception(msg)

    # ---------------
    # key
    # ---------------

    if lc[0]:

        key, key_cam = coll.get_diagnostic_cam(
            key=key,
            key_cam=key_cam,
            default='all',
        )

        doptics = coll.dobj['diagnostic'][key]['doptics']
        key_optics = list(itt.chain.from_iterable([
            [kcam] + [k0 for k0 in doptics[kcam]['optics']]
            for kcam in key_cam
        ]))

    # ---------------
    # key_optics
    # ---------------

    else:

        lok_ap = list(coll.dobj.get('aperture', {}).keys())
        lok_filt = list(coll.dobj.get('filter', {}).keys())
        lok_cryst = list(coll.dobj.get('crystal', {}).keys())
        lok_cam = list(coll.dobj.get('camera', {}).keys())

        if isinstance(key_optics, str):
            key_optics = [key_optics]

        key_optics = ds._generic_check._check_var_iter(
            key_optics, 'key_optics',
            types=(list, tuple, set),
            types_iter=str,
            allowed=lok_ap + lok_filt + lok_cryst + lok_cam,
        )

    # ---------------
    # factor
    # ---------------

    factor = float(ds._generic_check._check_var(
        factor, 'factor',
        types=(float, int),
        default=1.,
    ))

    # ---------------
    # iso
    # ---------------

    iso = 'ISO-10303-21'

    # ---------------
    # pfe_save
    # ---------------

    # Default
    if pfe_save is None:
        path = os.path.abspath('.')
        name = key if key is not None else _NAME
        pfe_save = os.path.join(path, f"{name}.{ext}")

    # check
    c0 = (
        isinstance(pfe_save, str)
        and (
            os.path.split(pfe_save)[0] == ''
            or os.path.isdir(os.path.split(pfe_save)[0])
        )
    )
    if not c0:
        msg = (
            f"Arg pfe_save must be a saving file str ending in '.{ext}'!\n"
            f"Provided: {pfe_save}"
        )
        raise Exception(msg)

    # makesure extension is included
    if not pfe_save.endswith(f'.{ext}'):
        pfe_save = f"{pfe_save}.{ext}"

    # ----------------
    # overwrite
    # ----------------

    overwrite = ds._generic_check._check_var(
        overwrite, 'overwrite',
        types=bool,
        default=False,
    )

    return (
        key_optics,
        factor, color,
        iso,
        pfe_save, overwrite,
    )


# #################################################################
# #################################################################
#          extract
# #################################################################


def _extract(
    coll=None,
    key_optics=None,
):

    # ----------------------
    # initialize
    # ----------------------

    dptsx = {}
    dptsy = {}
    dptsz = {}
    duvect = {}

    # ----------------------
    # extract points
    # ----------------------

    for i0, k0 in enumerate(key_optics):

        # points of polygons
        ptsx, ptsy, ptsz = coll.get_optics_poly(
            key=k0,
            add_points=False,
            # min_threshold=4e-3,
            min_threshold=None,
            mode=None,
            closed=True,
            ravel=None,
            total=True,
            return_outline=False,
        )

        # store
        if ptsx.ndim == 1:
            dptsx[k0], dptsy[k0], dptsz[k0] = ptsx, ptsy, ptsz

        elif ptsx.ndim == 2:
            for ii in range(ptsx.shape[1]):
                key = f"{k0}_{ii}"
                dptsx[key] = ptsx[:, ii]
                dptsy[key] = ptsy[:, ii]
                dptsz[key] = ptsz[:, ii]

        else:
            raise NotImplementedError(str(ptsx.shape))

        # unit vectors
        k0, cls = coll.get_optics_cls(optics=k0)
        k0, cls = k0[0], cls[0]
        if cls == 'camera':

            if coll.dobj[cls][k0]['dgeom']['parallel'] is True:
                duvect[k0] = {
                    'nin': coll.dobj[cls][k0]['dgeom']['nin'],
                    'e0': coll.dobj[cls][k0]['dgeom']['e0'],
                    'e1': coll.dobj[cls][k0]['dgeom']['e1'],
                }

            else:
                dv = coll.get_camera_unit_vectors(k0)
                lv = [
                    'nin_x', 'nin_y', 'nin_z',
                    'e0_x', 'e0_y', 'e0_z',
                    'e1_x', 'e1_y', 'e1_z',
                ]
                nin_x, nin_y, nin_z, e0x, e0y, e0z, e1x, e1y, e1z = [
                    dv[k1] for k1 in lv
                ]

                if e0x.ndim == 1:
                    for ii in range(ptsx.shape[1]):
                        key = f"{k0}_{ii}"
                        duvect[key] = {
                            'nin': np.r_[nin_x[ii], nin_y[ii], nin_z[ii]],
                            'e0': np.r_[e0x[ii], e0y[ii], e0z[ii]],
                            'e1': np.r_[e1x[ii], e1y[ii], e1z[ii]],
                        }

                else:
                    raise NotImplementedError(str(e0x.shape))

        else:
            duvect[k0] = {
                'nin': coll.dobj[cls][k0]['dgeom']['nin'],
                'e0': coll.dobj[cls][k0]['dgeom']['e0'],
                'e1': coll.dobj[cls][k0]['dgeom']['e1'],
            }

    # ------------------------
    # correspondence pts vect

    dcor_ptsvect = {}
    for k0 in dptsx.keys():
        if k0 in duvect.keys():
            dcor_ptsvect[k0] = k0
        else:
            key = '_'.join(k0.split('_')[:-1])
            assert key in duvect.keys(), (key, duvect.keys())
            dcor_ptsvect[k0] = key

    return dptsx, dptsy, dptsz, duvect, dcor_ptsvect


# #################################################################
# #################################################################
#          DATA
# #################################################################


def _get_data(
    dptsx=None,
    dptsy=None,
    dptsz=None,
    duvect=None,
    dcor_ptsvect=None,
    fname=None,
    # options
    dcolor=None,
    # norm
    iso=None,
    # empty names
    empty_name=None,
):

    empty_name = ds._generic_check._check_var(
        empty_name, 'empty_name',
        types=bool,
        default=False,
    )

    # -----------
    # npts
    # -----------

    # vectors
    dvx = {k0: np.diff(v0, axis=0) for k0, v0 in dptsx.items()}
    dvy = {k0: np.diff(v0, axis=0) for k0, v0 in dptsy.items()}
    dvz = {k0: np.diff(v0, axis=0) for k0, v0 in dptsz.items()}

    # dok
    dok = {k0: np.isfinite(v0) for k0, v0 in dvx.items()}

    # length
    dlength = {
        k0: np.sqrt(dvx[k0]**2 + dvy[k0]**2 + dvz[k0]**2)
        for k0 in dptsx.keys()
    }

    # directions
    ddx = {k0: dvx[k0] / dlength[k0] for k0 in dptsx.keys()}
    ddy = {k0: dvy[k0] / dlength[k0] for k0 in dptsx.keys()}
    ddz = {k0: dvz[k0] / dlength[k0] for k0 in dptsx.keys()}

    # shapes
    # dshape_vect = {k0: dvx[k0].shape for k0 in dptsx.keys()}

    dnpts = {k0: v0.sum() for k0, v0 in dok.items()}
    npts = np.sum([v0 for v0 in dnpts.values()])
    nsurf = len(dnpts)

    # ---------------
    # order of optics

    lksort = sorted(dptsx.keys())
    iss = np.argsort([dnpts[k0] for k0 in lksort])[::-1]
    lksort = [lksort[ii] for ii in iss]
    k0ind = _class02_save2stp._get_k0ind(
        dind_ok={k0: v0.nonzero() for k0, v0 in dok.items()},
        ncum=np.cumsum([dnpts[ksurf] for ksurf in lksort]),
        lkcam=lksort,
    )

    # index of first point cumulated
    dind_surf, i0 = {}, 0
    for ii, ksurf in enumerate(lksort):
        dind_surf[ksurf] = np.arange(0, dnpts[ksurf]) + i0
        i0 += dnpts[ksurf]

    # -----------
    # planes (can be less than optics)
    # -----------

    nplanes = len(duvect)
    lkplanes = sorted(duvect.keys())

    # -----------
    # colors
    # -----------

    colors = sorted(set([v0 for v0 in dcolor.values()]))
    ncol = len(colors)

    # -----------------
    # get index
    # ------------------

    i0 = 31
    dind = {
        'PRESENTATION_LAYER_ASSIGNMENT': {'order': 1},
        'STYLED_ITEM': {
            'order': 2,
            'nn': nsurf,
        },
        'PRESENTATION_STYLE_ASSIGNMENT': {
            'order': 3,
            'nn': ncol, # nsurf, #ncol,
        },
        'SURFACE_STYLE_USAGE': {
            'order': 4,
            'nn': ncol, # nsurf, #ncol,
        },
        'SURFACE_SIDE_STYLE': {
            'order': 5,
            'nn': ncol, #nsurf, #ncol,
        },
        'SURFACE_STYLE_FILL_AREA': {
            'order': 6,
            'nn': ncol, # nsurf, #ncol,
        },
        'FILL_AREA_STYLE': {
            'order': 7,
            'nn': ncol, # nsurf, #ncol,
        },
        'FILL_AREA_STYLE_COLOUR': {
            'order': 8,
            'nn': ncol, # nsurf, #ncol,
        },
        'COLOUR_RGB': {
            'order': 9,
            'nn': ncol,
        },
        'SHELL_BASED_SURFACE_MODEL': {
            'order': 10,
            'nn': nsurf,
        },
        'OPEN_SHELL': {
            'order': 11,
            'nn': nsurf,
        },
        'ADVANCED_FACE': {
            'order': 12,
            'nn': nsurf,
        },
        'PLANE': {
            'order': 13,
            'nn': nsurf,
        },
        'FACE_OUTER_BOUND': {
            'order': 14,
            'nn': nsurf,
        },
        'EDGE_LOOP': {
            'order': 15,
            'nn': nsurf,
        },
        'ORIENTED_EDGE': {
            'order': 16,
            'nn': npts,
        },
        'VERTEX_POINT': {
            'order': 17,
            'nn': npts,
        },
        'EDGE_CURVE': {
            'order': 18,
            'nn': npts,
        },
        'LINE': {
            'order': 19,
            'nn': npts,
        },
        'VECTOR': {
            'order': 20,
            'nn': npts,
        },
        'AXIS2_PLACEMENT_3D0': {
            'order': 21,
        },
        'AXIS2_PLACEMENT_3D': {
            'order': 22,
            'nn': nsurf,
        },
        'DIRECTION0': {
            'order': 23,
            'str': "DIRECTION('',(0.,0.,1.));",
        },
        'DIRECTION1': {
            'order': 24,
            'str': "DIRECTION('',(1.,0.,0.));",
        },
        'DIRECTION': {
            'order': 25,
            'nn': npts,
        },
        'DIRECTION_PLANES0': {
            'order': 26,
            'nn': nplanes,
        },
        'DIRECTION_PLANES1': {
            'order': 27,
            'nn': nplanes,
        },
        'CARTESIAN_POINT0': {
            'order': 28,
            'str': "CARTESIAN_POINT('',(0.,0.,0.));",
        },
        'CARTESIAN_POINT': {
            'order': 29,
            'nn': npts, # 2*(npts+1),
        },
        'MECHANICAL_DESIGN_GEOMETRIC_PRESENTATION_REPRESENTATION': {'order': 30},
    }

    # complement
    lkey = [k0 for k0 in dind.keys()]
    lorder = [dind[k0]['order'] for k0 in lkey]

    # safety ceck
    assert np.unique(lorder).size == len(lorder)
    inds = np.argsort(lorder)
    lkey = [lkey[ii] for ii in inds]

    # derive indices
    for k0 in lkey:
        nn = dind[k0].get('nn', 1)
        dind[k0]['ind'] = i0 + np.arange(0, nn)
        i0 += nn

    # -----------------
    # CARTESIAN_POINT
    # -----------------

    k0 = 'CARTESIAN_POINT'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        ksurf, ind = k0ind(ii)
        name = '' if empty_name else f'{ksurf}_{ind}'
        lines.append(f"#{ni}={k0}('{name}',({dptsx[ksurf][ind]},{dptsy[ksurf][ind]},{dptsz[ksurf][ind]}));")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # DIRECTION
    # -----------------

    k0 = 'DIRECTION'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        ksurf, ind = k0ind(ii)
        name = '' if empty_name else f'{ksurf}_{ind}'
        lines.append(f"#{ni}={k0}('{name}',({ddx[ksurf][ind]},{ddy[ksurf][ind]},{ddz[ksurf][ind]}));")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # DIRECTION_PLANES0
    # -----------------

    k0 = 'DIRECTION_PLANES0'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        kplane = lkplanes[ii]
        name = '' if empty_name else f'{kplane}'
        lines.append(f"#{ni}=DIRECTION('{name}',({duvect[kplane]['nin'][0]},{duvect[kplane]['nin'][1]},{duvect[kplane]['nin'][2]}));")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # DIRECTION_PLANES1
    # -----------------

    k0 = 'DIRECTION_PLANES1'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        kplane = lkplanes[ii]
        name = '' if empty_name else f'{kplane}'
        lines.append(f"#{ni}=DIRECTION('{name}',({duvect[kplane]['e0'][0]},{duvect[kplane]['e0'][1]},{duvect[kplane]['e0'][2]}));")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # VECTOR
    # -----------------

    k0 = 'VECTOR'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        ksurf, ind = k0ind(ii)
        name = '' if empty_name else f'{ksurf}_{ind}'
        lines.append(f"#{ni}={k0}('{name}',#{dind['DIRECTION']['ind'][ii]},1.);")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # AXIS2_PLACEMENT_3D0
    # -----------------

    k0 = 'AXIS2_PLACEMENT_3D0'
    ni = dind[k0]['ind'][0]
    dind[k0]['msg'] = f"#{ni}={k0[:-1]}('',#{dind['CARTESIAN_POINT0']['ind'][0]},#{dind['DIRECTION0']['ind'][0]},#{dind['DIRECTION1']['ind'][0]});"

    # -----------------
    # AXIS2_PLACEMENT_3D - TBF
    # -----------------

    k0 = 'AXIS2_PLACEMENT_3D'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        ksurf = lksort[ii]
        kplane = dcor_ptsvect[ksurf]
        iplane = lkplanes.index(kplane)
        ipt = dind['CARTESIAN_POINT']['ind'][dind_surf[ksurf][0]]
        # ipt = naxis[ii]
        lines.append(f"#{ni}={k0}('',#{ipt},#{dind['DIRECTION_PLANES0']['ind'][iplane]},#{dind['DIRECTION_PLANES1']['ind'][iplane]});")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # LINE
    # -----------------

    k0 = 'LINE'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        ksurf, ind = k0ind(ii)
        name = '' if empty_name else f'{ksurf}_{ind}'
        lines.append(f"#{ni}={k0}('{name}',#{dind['CARTESIAN_POINT']['ind'][ii]},#{dind['VECTOR']['ind'][ii]});")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # VERTEX_POINT
    # -----------------

    k0 = 'VERTEX_POINT'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        ksurf, ind = k0ind(ii)
        name = '' if empty_name else f'{ksurf}_{ind}'
        lines.append(f"#{ni}={k0}('{name}',#{dind['CARTESIAN_POINT']['ind'][ii]});")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # EDGE_CURVE
    # -----------------

    k0 = 'EDGE_CURVE'
    lines = []
    nn = np.array([dnpts[ksurf] for ksurf in lksort])
    icum = np.cumsum(nn) - 1
    nind = np.arange(0, dind[k0]['ind'].size) + 1
    nind[icum] -= nn
    for ii, ni in enumerate(dind[k0]['ind']):
        ksurf, ind = k0ind(ii)
        i1 = nind[ii]
        name = '' if empty_name else f'{ksurf}_{ind}'
        lines.append(f"#{ni}={k0}('{name}',#{dind['VERTEX_POINT']['ind'][ii]},#{dind['VERTEX_POINT']['ind'][i1]},#{dind['LINE']['ind'][ii]},.T.);")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # ORIENTED_EDGE
    # -----------------

    k0 = 'ORIENTED_EDGE'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        ksurf, ind = k0ind(ii)
        name = '' if empty_name else f'{ksurf}_{ind}'
        lines.append(f"#{ni}={k0}('{name}',*,*,#{dind['EDGE_CURVE']['ind'][ii]},.T.);")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # EDGE_LOOP
    # -----------------

    k0 = 'EDGE_LOOP'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        ksurf = lksort[ii]
        lstr = ",".join([f"#{dind['ORIENTED_EDGE']['ind'][jj]}" for jj in dind_surf[ksurf]])
        name = '' if empty_name else f'{ksurf}'
        lines.append(f"#{ni}={k0}('{name}',({lstr}));")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # FACE_OUTER_BOUND
    # -----------------

    k0 = 'FACE_OUTER_BOUND'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        ksurf = lksort[ii]
        name = '' if empty_name else f'{ksurf}'
        lines.append(f"#{ni}={k0}('{name}',#{dind['EDGE_LOOP']['ind'][ii]},.T.);")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # PLANE
    # -----------------

    k0 = 'PLANE'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        ksurf = lksort[ii]
        name = '' if empty_name else f'{ksurf}'
        lines.append(f"#{ni}={k0}('{name}',#{dind['AXIS2_PLACEMENT_3D']['ind'][ii]});")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # ADVANCED_FACE
    # -----------------

    k0 = 'ADVANCED_FACE'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        ksurf = lksort[ii]
        name = '' if empty_name else f'{ksurf}'
        lines.append(f"#{ni}={k0}('{name}',(#{dind['FACE_OUTER_BOUND']['ind'][ii]}),#{dind['PLANE']['ind'][ii]},.T.);")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # OPEN_SHELL
    # -----------------

    k0 = 'OPEN_SHELL'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        ksurf = lksort[ii]
        name = '' if empty_name else f'{ksurf}'
        lines.append(f"#{ni}={k0}('{name}',(#{dind['ADVANCED_FACE']['ind'][ii]}));")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # OPEN_SHELL
    # -----------------

    k0 = 'SHELL_BASED_SURFACE_MODEL'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        ksurf = lksort[ii]
        name = '' if empty_name else f'{ksurf}'
        lines.append(f"#{ni}={k0}('{name}',(#{dind['OPEN_SHELL']['ind'][ii]}));")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # COLOUR_RGB
    # -----------------

    k0 = 'COLOUR_RGB'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        name = '' if empty_name else f'color {ii}'
        lines.append(f"#{ni}={k0}('{name}',{colors[ii][0]},{colors[ii][1]},{colors[ii][2]});")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # FILL_AREA_STYLE_COLOUR
    # -----------------

    k0 = 'FILL_AREA_STYLE_COLOUR'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        name = '' if empty_name else f'color {ii}'
        lines.append(f"#{ni}={k0}('{name}',#{dind['COLOUR_RGB']['ind'][ii]});")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # FILL_AREA_STYLE
    # -----------------

    k0 = 'FILL_AREA_STYLE'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        name = '' if empty_name else f'color {ii}'
        lines.append(f"#{ni}={k0}('{name}',(#{dind['FILL_AREA_STYLE_COLOUR']['ind'][ii]}));")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # SURFACE_STYLE_FILL_AREA
    # -----------------

    k0 = 'SURFACE_STYLE_FILL_AREA'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        lines.append(f"#{ni}={k0}(#{dind['FILL_AREA_STYLE']['ind'][ii]});")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # SURFACE_SIDE_STYLE
    # -----------------

    k0 = 'SURFACE_SIDE_STYLE'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        name = '' if empty_name else f'color {ii}'
        lines.append(f"#{ni}={k0}('{name}',(#{dind['SURFACE_STYLE_FILL_AREA']['ind'][ii]}));")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # SURFACE_STYLE_USAGE
    # -----------------

    k0 = 'SURFACE_STYLE_USAGE'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        lines.append(f"#{ni}={k0}(.BOTH.,#{dind['SURFACE_SIDE_STYLE']['ind'][ii]});")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # PRESENTATION_STYLE_ASSIGNMENT
    # -----------------

    k0 = 'PRESENTATION_STYLE_ASSIGNMENT'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        lines.append(f"#{ni}={k0}((#{dind['SURFACE_STYLE_USAGE']['ind'][ii]}));")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # STYLED_ITEM
    # -----------------

    k0 = 'STYLED_ITEM'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        ksurf = lksort[ii]
        jj = colors.index(dcolor[ksurf])
        name = '' if empty_name else f'{ksurf}'
        lines.append(f"#{ni}={k0}('{name}',(#{dind['PRESENTATION_STYLE_ASSIGNMENT']['ind'][jj]}),#{dind['SHELL_BASED_SURFACE_MODEL']['ind'][ii]});")
    dind[k0]['msg'] = "\n".join(lines)

    # ----------------------
    # PRESENTATION_LAYER_ASSIGNMENT
    # ----------------------

    k0 = 'PRESENTATION_LAYER_ASSIGNMENT'
    ni = dind[k0]['ind'][0]
    lstr = ','.join([f"#{ii}" for ii in dind['SHELL_BASED_SURFACE_MODEL']['ind']])
    dind[k0]['msg'] = f"#{ni}={k0}('1','Layer 1',({lstr}));"

    # ----------------------
    # MECHANICAL_DESIGN_GEOMETRIC_PRESENTATION_REPRESENTATION
    # ----------------------

    k0 = 'MECHANICAL_DESIGN_GEOMETRIC_PRESENTATION_REPRESENTATION'
    ni = dind[k0]['ind'][0]
    lstr = ','.join([f"#{ii}" for ii in dind['STYLED_ITEM']['ind']])
    dind[k0]['msg'] = f"#{ni}={k0}('',({lstr}),#{i0});"

    # ------------
    # LEFTOVERS
    # ------------

    for k0, v0 in dind.items():
        if v0.get('msg') is None:
            if v0.get('str') is None:
                msg = f"Looks like '{k0}' is missing!"
                raise Exception(msg)
            else:
                ni = dind[k0]['ind'][0]
                dind[k0]['msg'] = f"#{ni}={v0['str']}"

    # ---------------
    # update index
    # ---------------

    # 5->91
    ind = i0 + np.arange(0, 8)

    # --------------------
    # msg_pre
    # --------------------

    lstr0 = ','.join([f"#{ii}" for ii in dind['SHELL_BASED_SURFACE_MODEL']['ind']])

    msg_pre = (
f"""
DATA;
#10=PROPERTY_DEFINITION_REPRESENTATION(#14,#12);
#11=PROPERTY_DEFINITION_REPRESENTATION(#15,#13);
#12=REPRESENTATION('',(#16),#{ind[0]});
#13=REPRESENTATION('',(#17),#{ind[0]});
#14=PROPERTY_DEFINITION('pmi validation property','',#21);
#15=PROPERTY_DEFINITION('pmi validation property','',#21);
#16=VALUE_REPRESENTATION_ITEM('number of annotations',COUNT_MEASURE(0.));
#17=VALUE_REPRESENTATION_ITEM('number of views',COUNT_MEASURE(0.));
#18=SHAPE_REPRESENTATION_RELATIONSHIP('None','relationship between A_objects-None and A_objects-None',#30,#19);
#19=MANIFOLD_SURFACE_SHAPE_REPRESENTATION('A_objects-None',({lstr0}),#{ind[0]});
#20=SHAPE_DEFINITION_REPRESENTATION(#21,#30);
#21=PRODUCT_DEFINITION_SHAPE('','',#22);
#22=PRODUCT_DEFINITION(' ','',#24,#23);
#23=PRODUCT_DEFINITION_CONTEXT('part definition',#29,'design');
#24=PRODUCT_DEFINITION_FORMATION_WITH_SPECIFIED_SOURCE(' ',' ',#26, .NOT_KNOWN.);
#25=PRODUCT_RELATED_PRODUCT_CATEGORY('part','',(#26));
#26=PRODUCT('A_objects','A_objects',' ',(#27));
#27=PRODUCT_CONTEXT(' ',#29,'mechanical');
#28=APPLICATION_PROTOCOL_DEFINITION('international standard','automotive_design',2010,#29);
#29=APPLICATION_CONTEXT('core data for automotive mechanical design processes');
#30=SHAPE_REPRESENTATION('A_objects-None',(#{dind['AXIS2_PLACEMENT_3D0']['ind'][0]}),#{ind[0]});
"""
    )

    # --------------------
    # msg_post
    # --------------------

    msg_post = (
f"""
#{ind[0]}=(
GEOMETRIC_REPRESENTATION_CONTEXT(3)
GLOBAL_UNCERTAINTY_ASSIGNED_CONTEXT((#{ind[1]}))
GLOBAL_UNIT_ASSIGNED_CONTEXT((#{ind[7]},#{ind[3]},#{ind[2]}))
REPRESENTATION_CONTEXT('{fname}','TOP_LEVEL_ASSEMBLY_PART')
);
#{ind[1]}=UNCERTAINTY_MEASURE_WITH_UNIT(LENGTH_MEASURE(2.E-5),#{ind[7]},'DISTANCE_ACCURACY_VALUE','Maximum Tolerance applied to model');
#{ind[2]}=(
NAMED_UNIT(*)
SI_UNIT($,.STERADIAN.)
SOLID_ANGLE_UNIT()
);
#{ind[3]}=(
CONVERSION_BASED_UNIT('DEGREE',#{ind[5]})
NAMED_UNIT(#{ind[4]})
PLANE_ANGLE_UNIT()
);
#{ind[4]}=DIMENSIONAL_EXPONENTS(0.,0.,0.,0.,0.,0.,0.);
#{ind[5]}=PLANE_ANGLE_MEASURE_WITH_UNIT(PLANE_ANGLE_MEASURE(0.0174532925),#{ind[6]});
#{ind[6]}=(
NAMED_UNIT(*)
PLANE_ANGLE_UNIT()
SI_UNIT($,.RADIAN.)
);
#{ind[7]}=(
LENGTH_UNIT()
NAMED_UNIT(*)
SI_UNIT(.MILLI.,.METRE.)
);
ENDSEC;
END-{iso};"""
    )

    # --------------------
    # assemble
    # --------------------

    msg = msg_pre + "\n".join([dind[k0]['msg'] for k0 in lkey]) + msg_post

    return msg


# -*- coding: utf-8 -*-
"""
Created on Wed Mar  8 11:00:59 2023

@author: dvezinet
"""

import warnings
import itertools as itt


import numpy as np


# ########################################################
# ########################################################
#               ref_vector_common
# ########################################################


def _get_ref_vector_common(
    coll=None,
    ddata=None,
    key_matrix=None,
    key_profile2d=None,
    dconstraints=None,
    strategy=None,
    strategy_bounds=None,
    # exclude from search
    key_exclude=None,
    ref_exclude=None,
    # others
    dref_vector=None,
):

    # ------------
    # check inputs

    lc = [
        ddata is not None,
        key_profile2d is not None
    ]
    if key_matrix is None or np.sum(lc) != 1:
        msg = "Please provide key_matrix and (ddata or key_profile2d)!"
        raise Exception(msg)

    if dref_vector is None:
        dref_vector = {}

    # ------------
    # geom matrix

    refc0 = None
    wbs = coll._which_bsplines
    key_bs = coll.dobj['geom matrix'][key_matrix]['bsplines']
    key_cam = coll.dobj['geom matrix'][key_matrix]['camera']
    refbs = coll.dobj[wbs][key_bs]['ref_bs']
    lgeom = coll.dobj['geom matrix'][key_matrix]['data']
    for ii, k0 in enumerate(lgeom):
        camdgeom = coll.dobj['camera'][key_cam[ii]]['dgeom']
        camref = camdgeom['ref'] + camdgeom['ref_flat']
        refi = [
            rr for rr in coll.ddata[k0]['ref']
            if rr not in camref
            and rr not in refbs
        ]

        # safety check
        if len(refi) > 1:
            msg = (
                "Geometry matrix with > 1 extra ref not handled yet\n"
                f"\t- gome. mat: '{key_matrix}'\n"
                f"\t- ref: {coll.ddata[k0]['ref']}"
            )
            raise NotImplementedError(msg)

        # sort cases
        if len(refi) == 0:
            assert refc0 is None

        elif len(refi) == 1:
            if refc0 is None:
                assert ii == 0
                refc0 = refi[0]

            assert refc0 == refi[0]

    lk = list(coll.dobj['geom matrix'][key_matrix]['data'])

    # ------------------
    # data or profile2d

    refc1 = None
    if ddata is not None:

        # handle equivalent diag / cameras
        kcameq = 'keys_cam' if ddata['keys_cam_equi'] is None else 'keys_cam_equi'
        key_cam = ddata[kcameq]

        # loop
        for ii, k0 in enumerate(ddata['keys']):
            refi = [
                rr for rr in coll.ddata[k0]['ref']
                if rr not in coll.dobj['camera'][key_cam[ii]]['dgeom']['ref']
            ]

            if len(refi) > 1:
                msg = (
                    "Multiple possible non-camera refs for ddata[k0]:\n"
                    f"\t- k0: {k0}\n"
                    f"\t- ii: {ii}\n"
                    f"\t- key_cam_equi: {ddata['keys_cam'] is not None}\n"
                    f"\t- key_cam[ii]: {key_cam[ii]}\n"
                    f"\t- coll.dobj['camera']['{key_cam[ii]}']['dgeom']['ref']: {coll.dobj['camera'][key_cam[ii]]['dgeom']['ref']}\n"
                    f"\t- coll.ddata['{k0}']['ref']: {coll.ddata[k0]['ref']}\n"
                    f"\t- refi: {refi}\n"
                    f"\t- key_cam: {key_cam}\n"
                    f"\t- key_matrix: {key_matrix}\n"
                    f"\t- key_profile2d: {key_profile2d}\n"
                    f"\t- strategy: {strategy}\n"
                )
                raise Exception(msg)

            if len(refi) == 0:
                assert refc1 is None
            elif len(refi) == 1:
                if refc1 is None:
                    assert ii == 0
                    refc1 = refi[0]
                assert refc1 == refi[0]

        lk += ddata['keys']

        # ------------
        # dconstraints

        if dconstraints is not None:
            if isinstance(dconstraints.get('rmax', {}).get('val'), str):
                lk.append(dconstraints['rmax']['val'])
            if isinstance(dconstraints.get('rmin', {}).get('val'), str):
                lk.append(dconstraints['rmin']['val'])

    # ----------
    # profile2d

    else:

        lrefbs = list(itt.chain.from_iterable([
            v0['ref'] for v0 in coll.dobj[wbs].values()
        ]))

        key_bs = coll.get_profiles2d()[key_profile2d]
        refbs = coll.dobj[wbs][key_bs]['ref_bs']
        refi = [
            rr for rr in coll.ddata[key_profile2d]['ref']
            if rr not in refbs
            # and (refc0 is None or rr == refc0)
            and rr not in lrefbs
        ]
        if len(refi) > 1:
            msg = (key_profile2d, coll.ddata[key_profile2d]['ref'], refi, lrefbs)
            raise Exception(msg)

        if len(refi) == 1:
            refc1 = refi[0]

        lk += [key_profile2d]

    # -----
    # refc

    if refc0 is None and refc1 is None:
        return False, None, None, None, None

    else:

        if refc0 is not None and refc1 is not None:

            if refc0 != refc1:
                msg = (
                    "Non-consistent references with:\n"
                    f"\t- dref_vector: {dref_vector}\n"
                    "For:\n"
                    f"\t- geom matrix: {refc0}\n"
                )
                if ddata is None:
                    msg += f"\t- {key_profile2d}: {refc1}\n"
                else:
                    msg += f"\t- {ddata['keys']}: {refc1}\n"
                warnings.warn(msg)
                refc = None

            else:
                refc = refc0

        elif refc0 is None:
            refc = refc1

        else:
            refc = refc0

        return coll.get_ref_vector_common(
            keys=lk,
            ref=refc,
            strategy=strategy,
            strategy_bounds=strategy_bounds,
            # exclude from search
            key_exclude=key_exclude,
            ref_exclude=ref_exclude,
            # others
            **dref_vector,
        )


# -*- coding: utf-8 -*-


import numpy as np
import datastock as ds


from ..geom._comp_solidangles import _check_polygon_2d, _check_polygon_3d


# #################################################################
# #################################################################
#               Surface 3d
# #################################################################


def _surface3d(
    key=None,
    # 2d outline
    outline_x0=None,
    outline_x1=None,
    cent=None,
    # 3d outline
    poly_x=None,
    poly_y=None,
    poly_z=None,
    # extenthalf
    extenthalf=None,
    # normal vector at cent
    nin=None,
    e0=None,
    e1=None,
    # curvature
    curve_r=None,
    make_planar=None,
):

    # -----------
    # cent

    if cent is not None:
        cent = np.atleast_1d(cent).ravel().astype(float)
        assert cent.shape == (3,)

    # -----------
    # unit vectors

    # nin has to be provided
    nin = ds._generic_check._check_flat1darray(
        var=nin, varname='nin', dtype=float, size=3, norm=True,
    )

    nin, e0, e1 = ds._generic_check._check_vectbasis(
        e0=nin,
        e1=e0,
        e2=e1,
        dim=3,
    )

    # -----------------------------
    # outline vs poly vs extenthalf

    lc = [
        all([pp is not None for pp in [outline_x0, outline_x1, cent, e0]]),
        all([pp is not None for pp in [extenthalf, curve_r, cent, e0]]),
        all([pp is not None for pp in [poly_x, poly_y, poly_z]])
    ]
    if np.sum(lc) != 1:
        kwd = locals()

        lk = [
            'outline_x0', 'outline_x1', 'cent', 'e0',
            'extenthalf', 'curve_r', 'cent', 'e0',
            'poly_x', 'poly_y', 'poly_z',
        ]
        lstr = [f"\t- {ss}: {kwd[ss]}" for ss in lk]
        msg = (
            "Please provide either (xor):\n"
            "\t- planar: outline_x0, outline_x1 and cent, e0, e1\n"
            "xor\n"
            "\t- curved: extenthalf and cent, e0, e1\n"
            "xor\n"
            "\t- arbitrary 3d: poly_x, poly_y, poly_z\nn"
            "Provided:\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)

    # --------------
    # outline

    if lc[0]:

        # check outline
        outline_x0, outline_x1, area = _check_polygon_2d(
            poly_x=outline_x0,
            poly_y=outline_x1,
            poly_name=f'{key}_outline',
            can_be_None=False,
            closed=False,
            counter_clockwise=True,
            return_area=True,
        )

        # if rectangle, try to derive extenthalf and r_curve
        if np.unique(outline_x0).size == 2:
            if np.unique(outline_x1).size == 2:
                dx0 = outline_x0.max() - outline_x0.min()
                dx1 = outline_x1.max() - outline_x1.min()
                extenthalf = [dx0*0.5, dx1*0.5]
                curve_r = [np.inf, np.inf]

        # derive poly 3d
        # poly_x = cent[0] + outline_x0 * e0[0] + outline_x1 * e1[0]
        # poly_y = cent[1] + outline_x0 * e0[1] + outline_x1 * e1[1]
        # poly_z = cent[2] + outline_x0 * e0[2] + outline_x1 * e1[2]

        gtype = 'planar'

    # ----------
    # curvature

    if lc[1]:

        curve_r = ds._generic_check._check_flat1darray(
            curve_r, 'curve_r', size=[1, 2], dtype=float,
        )
        extenthalf = ds._generic_check._check_flat1darray(
            extenthalf, 'extenthalf', size=[1, 2], dtype=float,
        )
        if extenthalf.size == 1:
            extenthalf = np.r_[extenthalf, extenthalf]

        if np.any(np.isnan(curve_r)):
            msg = (
                f"Arg curve_r for 3d surface '{key}' must:\n"
                "\t- be array-like of floats\n"
                "\t- be of size <= 2\n"
                f"Provided: {curve_r}"
            )
            raise Exception(msg)

        if curve_r.size == 1:
            curve_r = np.r_[curve_r, curve_r]

        ninf = np.isinf(curve_r).sum()
        if ninf == 1:
            gtype = 'cylindrical'
        elif ninf == 2:
            gtype = 'planar'
        elif np.unique(curve_r).size == 1:
            gtype = 'spherical'
        else:
            gtype = 'toroidal'

        curve_r = tuple(curve_r)

        # area
        area = _get_curved_area(
            gtype=gtype,
            curve_r=np.abs(curve_r),
            extenthalf=extenthalf,
        )

        # outline in local coordinates
        outline_x0 = extenthalf[0]*np.r_[-1, 1, 1, -1]
        outline_x1 = extenthalf[1]*np.r_[-1, -1, 1, 1]

    # --------------------------------------------
    # try to get 2d outline from 3d poly if planar

    if lc[2]:

        # poly 3d sanity check
        poly_x, poly_y, poly_z = _check_polygon_3d(
            poly_x=poly_x,
            poly_y=poly_y,
            poly_z=poly_z,
            poly_name=f'{key}-polygon',
            can_be_None=False,
            closed=False,
            counter_clockwise=True,
            normal=nin,
        )

        gtype, cent, outline_x0, outline_x1, area = _get_outline_from_poly(
            key=key,
            cent=cent,
            poly_x=poly_x,
            poly_y=poly_y,
            poly_z=poly_z,
            nin=nin,
            e0=e0,
            e1=e1,
            make_planar=make_planar,
        )

        if outline_x0 is not None:
            poly_x, poly_y, poly_z = None, None, None

    # ----------
    # return

    return (
        cent,
        outline_x0, outline_x1,
        poly_x, poly_y, poly_z,
        nin, e0, e1,
        extenthalf, area, curve_r, gtype,
    )


# #################################################################
# #################################################################
#               curved surfaces
# #################################################################


def _get_curved_area(
    gtype=None,
    curve_r=None,
    extenthalf=None,
):

    # planar
    if gtype == 'planar':
        area = 4. * extenthalf[0] * extenthalf[1]

    # cylindrical
    if gtype == 'cylindrical':
        iplan = np.isinf(curve_r).nonzero()[0][0]
        icurv = 1 - iplan
        rc = curve_r[icurv]

        area = 2. * extenthalf[iplan] * rc * 2. * extenthalf[icurv]

    # spherical
    elif gtype == 'spherical':
        rc = curve_r[0]
        ind = np.argmax(extenthalf)
        dphi = extenthalf[ind]
        sindtheta = np.sin(extenthalf[ind-1])

        area = 4. * rc**2 * dphi * sindtheta

    # toroidal
    elif gtype == 'toroidal':
        imax = np.argmax(curve_r)
        imin = 1 - imax
        rmax = curve_r[imax]
        rmin = curve_r[imin]
        phi2 = extenthalf[imax]
        theta2 = extenthalf[imin]

        area = rmin * 2.*phi2 * (rmax*2*theta2 + 2*rmin*np.sin(theta2))

    return area


def _get_curved_poly(
    gtype=None,
    curve_r=None,
    outline_x0=None,
    outline_x1=None,
    cent=None,
    nin=None,
    e0=None,
    e1=None,
):

    # ------------
    # compute

    # planar
    if gtype == 'planar':

        poly_x = cent[0] + outline_x0 * e0[0] + outline_x1 * e1[0]
        poly_y = cent[1] + outline_x0 * e0[1] + outline_x1 * e1[1]
        poly_z = cent[2] + outline_x0 * e0[2] + outline_x1 * e1[2]

    # cylindrical
    if gtype == 'cylindrical':
        iplan = np.isinf(curve_r).nonzero()[0][0]
        icurv = 1 - iplan
        rc = curve_r[icurv]
        rcs = np.sign(rc)
        rca = np.abs(rc)

        centbis = cent + rc * nin
        ee = [e0, e1]
        outline = [outline_x0, outline_x1]

        ang = outline[icurv]
        xx = outline[iplan]

        vx = np.cos(ang) * (-rcs*nin[0]) + np.sin(ang) * ee[icurv][0]
        vy = np.cos(ang) * (-rcs*nin[1]) + np.sin(ang) * ee[icurv][1]
        vz = np.cos(ang) * (-rcs*nin[2]) + np.sin(ang) * ee[icurv][2]

        poly_x = centbis[0] + xx * ee[iplan][0] + rca * vx
        poly_y = centbis[1] + xx * ee[iplan][1] + rca * vy
        poly_z = centbis[2] + xx * ee[iplan][2] + rca * vz

    # spherical
    elif gtype == 'spherical':

        rc = curve_r[0]
        rcs = np.sign(rc)
        rca = np.abs(rc)

        centbis = cent + rc * nin
        psi, dtheta = outline_x0, outline_x1

        vpsix = np.cos(psi) * (-rcs*nin[0]) + np.sin(psi) * e0[0]
        vpsiy = np.cos(psi) * (-rcs*nin[1]) + np.sin(psi) * e0[1]
        vpsiz = np.cos(psi) * (-rcs*nin[2]) + np.sin(psi) * e0[2]

        vx = np.cos(dtheta) * vpsix + np.sin(dtheta) * e1[0]
        vy = np.cos(dtheta) * vpsiy + np.sin(dtheta) * e1[1]
        vz = np.cos(dtheta) * vpsiz + np.sin(dtheta) * e1[2]

        poly_x = centbis[0] + rca * vx
        poly_y = centbis[1] + rca * vy
        poly_z = centbis[2] + rca * vz

    # toroidal
    elif gtype == 'toroidal':

        imax = np.argmax(np.abs(curve_r))
        imin = 1 - imax
        rmax = curve_r[imax]
        rmin = curve_r[imin]
        rmaxs = np.sign(rmax)
        rmins = np.sign(rmin)
        rmaxa = np.abs(rmax)
        rmina = np.abs(rmin)
        crosss = rmaxs*rmins

        cmax = cent + nin * (rmax + rmin)
        outline = [outline_x0, outline_x1]
        amax = outline[imax]
        amin = outline[imin]

        ee = [e0, e1]

        vmaxx = np.cos(amax) * (-rmaxs*nin)[0] + np.sin(amax) * ee[imax][0]
        vmaxy = np.cos(amax) * (-rmaxs*nin)[1] + np.sin(amax) * ee[imax][1]
        vmaxz = np.cos(amax) * (-rmaxs*nin)[2] + np.sin(amax) * ee[imax][2]

        cminx = cmax[0] + rmaxa * vmaxx
        cminy = cmax[1] + rmaxa * vmaxy
        cminz = cmax[2] + rmaxa * vmaxz

        vminx = np.cos(amin) * crosss * vmaxx + np.sin(amin) * ee[imin][0]
        vminy = np.cos(amin) * crosss * vmaxy + np.sin(amin) * ee[imin][1]
        vminz = np.cos(amin) * crosss * vmaxz + np.sin(amin) * ee[imin][2]

        poly_x = cminx + rmina * vminx
        poly_y = cminy + rmina * vminy
        poly_z = cminz + rmina * vminz

    return poly_x, poly_y, poly_z


# ###############################################################
# ###############################################################
#               outline from poly 3d
# ###############################################################


def _get_outline_from_poly(
    key=None,
    cent=None,
    poly_x=None,
    poly_y=None,
    poly_z=None,
    nin=None,
    e0=None,
    e1=None,
    make_planar=None,
):

    # --------
    # check input

    make_planar = ds._generic_check._check_var(
        make_planar, 'make_planar',
        types=bool,
        default=True,
    )

    if make_planar is True and e0 is None:
        msg = "Provide e0 to make planar!"
        raise Exception(msg)

    # ----------
    # cent

    if cent is None:
        cent = np.r_[np.mean(poly_x), np.mean(poly_y), np.mean(poly_z)]

    # ----------
    # planar

    if make_planar:
        # all deviation smaller than 1.e-10 degree
        gtype = 'planar'

        # derive outline
        outline_x0 = (
            (poly_x - cent[0]) * e0[0]
            + (poly_y - cent[1]) * e0[1]
            + (poly_z - cent[2]) * e0[2]
        )
        outline_x1 = (
            (poly_x - cent[0]) * e1[0]
            + (poly_y - cent[1]) * e1[1]
            + (poly_z - cent[2]) * e1[2]
        )

        # check outline
        outline_x0, outline_x1, area = _check_polygon_2d(
            poly_x=outline_x0,
            poly_y=outline_x1,
            poly_name=f'{key}-outline',
            can_be_None=False,
            closed=False,
            counter_clockwise=True,
            return_area=True,
        )

    else:
        gtype = '3d'
        outline_x0, outline_x1 = None, None
        area = np.nan

    return gtype, cent, outline_x0, outline_x1, area
